import json
from pathlib import Path

import httpx
import pytest
import respx

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.commands import UploadCommand
from cognite_toolkit._cdf_tk.storageio.selectors import FileMetadataTemplate, FileMetadataTemplateSelector
from cognite_toolkit._cdf_tk.storageio.selectors._file_content import FILENAME_VARIABLE


@pytest.fixture
def file_folder(tmp_path: Path) -> Path:
    # Create some dummy file content
    file_folder = tmp_path / "files"
    file_folder.mkdir(parents=True, exist_ok=True)
    for i in range(5):
        file_path = file_folder / f"file_{i}.txt"
        file_path.write_text(f"This is the content of file {i}.", encoding="utf-8")
    return file_folder


class TestFileContent:
    @pytest.mark.usefixtures("disable_gzip")
    def test_upload_filemetadata(
        self, respx_mock: respx.MockRouter, toolkit_config: ToolkitClientConfig, file_folder: Path
    ) -> None:
        config = toolkit_config
        client = ToolkitClient(config)
        file_upload_url = "https://upload.url/for/testing/{externalId}"

        def create_callback(request: httpx.Request) -> httpx.Response:
            payload = json.loads(request.content)
            assert "name" in payload
            assert "externalId" in payload
            payload["uploadUrl"] = file_upload_url.format(externalId=payload["externalId"])
            return httpx.Response(status_code=200, json=payload, headers={})

        respx_mock.post(config.create_api_url("/files")).mock(side_effect=create_callback)
        upload_endpoints: list[str] = []
        for filepath in file_folder.iterdir():
            if filepath.is_file():
                upload_endpoint = file_upload_url.format(externalId=f"my_file_{filepath.name}")
                respx_mock.put(upload_endpoint).mock(return_value=httpx.Response(status_code=200))
                upload_endpoints.append(upload_endpoint)

        selector = FileMetadataTemplateSelector(
            file_directory=file_folder,
            template=FileMetadataTemplate.model_validate(
                dict(
                    name=FILENAME_VARIABLE,
                    external_id=f"my_file_{FILENAME_VARIABLE}",
                    source="Uploaded via Toolkit",
                )
            ),
        )
        selector.dump_to_file(file_folder.parent)
        cmd = UploadCommand(silent=True)
        cmd.upload(
            input_dir=file_folder.parent,
            client=client,
            deploy_resources=False,
            dry_run=False,
            verbose=True,
        )

        assert respx_mock.calls.call_count == 10  # 5 for metadata, 5 for content
        called_endpoints = {str(call.request.url) for call in respx_mock.calls}
        assert set(upload_endpoints).issubset(called_endpoints)



from collections.abc import Hashable
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from cognite.client.data_classes import data_modeling as dm

from cognite_toolkit._cdf_tk.cruds import ContainerCRUD, ResourceCRUD, ResourceWorker, SpaceCRUD, ViewCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitCycleError
from tests.test_unit.approval_client import ApprovalToolkitClient


def _create_test_container(
    external_id: str,
    properties: dict[str, dm.ContainerProperty],
    constraints: dict | None = None,
    space: str = "my_space",
) -> dm.Container:
    """Helper to create a Container with standard test defaults."""
    return dm.Container(
        space=space,
        external_id=external_id,
        properties=properties,
        last_updated_time=1,
        created_time=1,
        is_global=False,
        used_for="node",
        constraints=constraints or {},
        description=None,
        name=None,
        indexes={},
    )


def _create_test_view(external_id: str, container: dm.Container) -> dm.View:
    """Helper to create a View that maps to all properties of the given Container."""
    properties = {}
    for prop_name, container_prop in container.properties.items():
        properties[prop_name] = dm.MappedProperty(
            container=container.as_id(),
            container_property_identifier=prop_name,
            type=container_prop.type,
            nullable=container_prop.nullable,
            immutable=container_prop.immutable,
            auto_increment=container_prop.auto_increment,
        )

    return dm.View(
        space=container.space,
        external_id=external_id,
        version="v1",
        properties=properties,
        last_updated_time=1,
        created_time=1,
        is_global=False,
        used_for="node",
        writable=True,
        description=None,
        name=None,
        filter=None,
        implements=None,
    )


class TestViewLoader:
    def test_unchanged_view_int_version(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        loader = ViewCRUD.create_loader(toolkit_client_approval.mock_client)
        raw_file = """- space: sp_space
  externalId: my_view
  version: 1"""
        file = MagicMock(spec=Path)
        file.read_text.return_value = raw_file
        cdf_view = dm.View(
            space="sp_space",
            external_id="my_view",
            version="1",
            last_updated_time=1,
            created_time=1,
            description=None,
            name=None,
            filter=None,
            implements=None,
            writable=True,
            used_for="node",
            is_global=False,
            properties={},
        )

        toolkit_client_approval.append(dm.View, [cdf_view])

        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([file])
        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}

    @pytest.mark.parametrize(
        "item, expected",
        [
            pytest.param(
                {
                    "space": "sp_my_space",
                    "properties": {
                        "name": {
                            "container": {
                                "type": "container",
                                "space": "my_container_space",
                                "externalId": "my_container",
                            }
                        }
                    },
                },
                [
                    (SpaceCRUD, "sp_my_space"),
                    (ContainerCRUD, dm.ContainerId(space="my_container_space", external_id="my_container")),
                ],
                id="View with one container property",
            ),
            pytest.param(
                {
                    "space": "sp_my_space",
                    "properties": {
                        "toEdge": {
                            "source": {
                                "type": "view",
                                "space": "my_view_space",
                                "externalId": "my_view",
                                "version": "1",
                            },
                            "edgeSource": {
                                "type": "view",
                                "space": "my_other_view_space",
                                "externalId": "my_edge_view",
                                "version": "42",
                            },
                        }
                    },
                },
                [
                    (SpaceCRUD, "sp_my_space"),
                    (ViewCRUD, dm.ViewId(space="my_view_space", external_id="my_view", version="1")),
                    (ViewCRUD, dm.ViewId(space="my_other_view_space", external_id="my_edge_view", version="42")),
                ],
                id="View with one container property",
            ),
        ],
    )
    def test_get_dependent_items(self, item: dict, expected: list[tuple[type[ResourceCRUD], Hashable]]) -> None:
        actual = ViewCRUD.get_dependent_items(item)

        assert list(actual) == expected

    @pytest.mark.parametrize(
        "view_ids,ordering_constraints,test_description",
        [
            pytest.param(
                [
                    dm.ViewId("cdf_cdm", "CogniteAsset", "v1"),
                    dm.ViewId("cdf_cdm", "CogniteSourceable", "v1"),
                    dm.ViewId("cdf_cdm", "CogniteSourceSystem", "v1"),
                ],
                [
                    (
                        dm.ViewId("cdf_cdm", "CogniteSourceSystem", "v1"),
                        dm.ViewId("cdf_cdm", "CogniteSourceable", "v1"),
                    ),
                    (dm.ViewId("cdf_cdm", "CogniteSourceable", "v1"), dm.ViewId("cdf_cdm", "CogniteAsset", "v1")),
                ],
                "Transitive chain: CogniteSourceSystem -> CogniteSourceable -> CogniteAsset",
                id="transitive_chain",
            ),
            pytest.param(
                [
                    dm.ViewId("cdf_cdm", "CogniteActivity", "v1"),
                    dm.ViewId("cdf_cdm", "CogniteSourceable", "v1"),
                    dm.ViewId("cdf_cdm", "CogniteSchedulable", "v1"),
                    dm.ViewId("cdf_cdm", "CogniteSourceSystem", "v1"),
                ],
                [
                    (
                        dm.ViewId("cdf_cdm", "CogniteSourceSystem", "v1"),
                        dm.ViewId("cdf_cdm", "CogniteSourceable", "v1"),
                    ),
                    (dm.ViewId("cdf_cdm", "CogniteSchedulable", "v1"), dm.ViewId("cdf_cdm", "CogniteActivity", "v1")),
                ],
                "Multiple independent chains: (CogniteSourceSystem -> CogniteSourceable) and (CogniteSchedulable -> CogniteActivity)",
                id="multiple_independent_chains",
            ),
            pytest.param(
                [
                    dm.ViewId("cdf_cdm", "CogniteAsset", "v1"),
                    dm.ViewId("cdf_cdm", "CogniteAssetClass", "v1"),
                    dm.ViewId("cdf_cdm", "CogniteAssetType", "v1"),
                    dm.ViewId("cdf_cdm", "CogniteDescribable", "v1"),
                ],
                [
                    (dm.ViewId("cdf_cdm", "CogniteDescribable", "v1"), dm.ViewId("cdf_cdm", "CogniteAssetType", "v1")),
                    (dm.ViewId("cdf_cdm", "CogniteDescribable", "v1"), dm.ViewId("cdf_cdm", "CogniteAssetClass", "v1")),
                    (dm.ViewId("cdf_cdm", "CogniteAssetClass", "v1"), dm.ViewId("cdf_cdm", "CogniteAsset", "v1")),
                    (dm.ViewId("cdf_cdm", "CogniteAssetType", "v1"), dm.ViewId("cdf_cdm", "CogniteAsset", "v1")),
                ],
                "Diamond: CogniteDescribable -> CogniteAssetType/CogniteAssetClass -> CogniteAsset",
                id="diamond_dependency",
            ),
        ],
    )
    def test_topological_sort_container_constraints_dependency_patterns(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
        cognite_core_no_3D: dm.DataModel,
        cognite_core_containers_no_3D: dm.ContainerList,
        view_ids: list[dm.ViewId],
        ordering_constraints: list[tuple[dm.ViewId, dm.ViewId]],
        test_description: str,
    ) -> None:
        """Test various dependency patterns: transitive chains, independent chains, and diamond dependencies."""
        loader = ViewCRUD.create_loader(toolkit_client_approval.mock_client)
        toolkit_client_approval.append(dm.View, cognite_core_no_3D.views)
        toolkit_client_approval.append(dm.Container, cognite_core_containers_no_3D)

        sorted_views = loader.topological_sort_container_constraints(view_ids)

        # Verify same number of views returned
        assert len(sorted_views) == len(view_ids), (
            f"{test_description}: Expected {len(view_ids)} views, got {len(sorted_views)}"
        )

        # Verify all ordering constraints
        for before_view_id, after_view_id in ordering_constraints:
            before_idx = sorted_views.index(before_view_id)
            after_idx = sorted_views.index(after_view_id)
            assert before_idx < after_idx, (
                f"{test_description}: {before_view_id.external_id} should come before {after_view_id.external_id}"
            )

    def test_topological_sort_container_constraints_cyclical_dependency(
        self, toolkit_client_approval: ApprovalToolkitClient
    ) -> None:
        """Test that cyclical dependencies raise ToolkitCycleError."""

        loader = ViewCRUD.create_loader(toolkit_client_approval.mock_client)

        # Create three containers that form a cycle: A -> B -> C -> A
        container_a = _create_test_container(
            "ContainerA",
            {
                "refC": dm.ContainerProperty(
                    type=dm.DirectRelation(container=dm.ContainerId("my_space", "ContainerC")),
                    nullable=True,
                    immutable=False,
                    auto_increment=False,
                )
            },
        )
        container_b = _create_test_container(
            "ContainerB",
            {
                "refA": dm.ContainerProperty(
                    type=dm.DirectRelation(container=dm.ContainerId("my_space", "ContainerA")),
                    nullable=True,
                    immutable=False,
                    auto_increment=False,
                )
            },
        )
        container_c = _create_test_container(
            "ContainerC",
            {
                "name": dm.ContainerProperty(type=dm.Text(), nullable=True, immutable=False, auto_increment=False),
            },
            {
                "requiresB": dm.RequiresConstraint(require=dm.ContainerId("my_space", "ContainerB")),
            },
        )

        view_a = _create_test_view("ViewA", container_a)
        view_b = _create_test_view("ViewB", container_b)
        view_c = _create_test_view("ViewC", container_c)

        toolkit_client_approval.append(dm.View, [view_a, view_b, view_c])
        toolkit_client_approval.append(dm.Container, [container_a, container_b, container_c])

        view_ids = [view_a.as_id(), view_b.as_id(), view_c.as_id()]

        with pytest.raises(ToolkitCycleError):
            loader.topological_sort_container_constraints(view_ids)

    @pytest.mark.parametrize(
        "view_id,expected_readonly_props",
        [
            pytest.param(
                dm.ViewId("cdf_cdm", "CogniteAsset", "v1"),
                {"pathLastUpdatedTime", "path", "root"},
                id="CogniteAsset_has_readonly_properties",
            ),
            pytest.param(
                dm.ViewId("cdf_cdm", "CogniteFile", "v1"),
                {"isUploaded", "uploadedTime"},
                id="CogniteFile_has_readonly_properties",
            ),
            pytest.param(
                dm.ViewId("cdf_cdm", "CogniteSourceSystem", "v1"),
                set(),
                id="CogniteSourceSystem_no_readonly_properties",
            ),
        ],
    )
    def test_get_readonly_properties(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
        cognite_core_no_3D: dm.DataModel,
        view_id: dm.ViewId,
        expected_readonly_props: set[str],
    ) -> None:
        """Test that get_readonly_properties identifies readonly properties from containers."""
        loader = ViewCRUD.create_loader(toolkit_client_approval.mock_client)
        toolkit_client_approval.append(dm.View, cognite_core_no_3D.views)

        readonly_props = loader.get_readonly_properties(view_id)
        assert readonly_props == expected_readonly_props



from typing import Literal

from cognite_toolkit._cdf_tk.constants import StreamTemplateName
from cognite_toolkit._cdf_tk.protocols import (
    ResourceRequestListProtocol,
    ResourceResponseListProtocol,
)

from .base import BaseModelObject, BaseResourceList, RequestResource, ResponseResource


class StreamRequest(RequestResource):
    """Stream request resource class."""

    external_id: str
    settings: dict[Literal["template"], dict[Literal["name"], StreamTemplateName]]

    def as_id(self) -> str:
        return self.external_id


class StreamRequestList(BaseResourceList[StreamRequest], ResourceRequestListProtocol):
    """List of Stream request resources."""

    _RESOURCE = StreamRequest


class LifecycleObject(BaseModelObject):
    """Lifecycle object."""

    hot_phase_duration: str | None = None
    data_deleted_after: str | None = None
    retained_after_soft_delete: str


class ResourceUsage(BaseModelObject):
    """Resource quota with provisioned and consumed values."""

    provisioned: int
    consumed: int | None = None


class LimitsObject(BaseModelObject):
    """Limits object."""

    max_records_total: ResourceUsage
    max_giga_bytes_total: ResourceUsage
    max_filtering_interval: str | None = None


class StreamSettings(BaseModelObject):
    """Stream settings object."""

    lifecycle: LifecycleObject
    limits: LimitsObject


class StreamResponse(ResponseResource["StreamRequest"]):
    """Stream response resource class."""

    external_id: str
    created_time: int
    created_from_template: StreamTemplateName
    type: Literal["Mutable", "Immutable"]
    settings: StreamSettings | None = None

    def as_request_resource(self) -> StreamRequest:
        return StreamRequest.model_validate(
            {
                "externalId": self.external_id,
                "settings": {"template": {"name": self.created_from_template}},
            }
        )

    def as_write(self) -> StreamRequest:
        return StreamRequest.model_validate(
            {
                "externalId": self.external_id,
                "settings": {"template": {"name": self.created_from_template}},
            }
        )


class StreamResponseList(BaseResourceList[StreamResponse], ResourceResponseListProtocol):
    """List of Stream response resources."""

    _RESOURCE = StreamResponse

    def as_write(self) -> StreamRequestList:
        return StreamRequestList([item.as_write() for item in self.data])



from typing import Annotated

import typer

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.commands import InitCommand
from cognite_toolkit._cdf_tk.feature_flags import Flags


class LandingApp(typer.Typer):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)

    def main_init(
        self,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run, do dry-run if present.",
            ),
        ] = False,
        # TODO: this is a temporary solution to be able to test the functionality
        # in a new environment, assuming that the toml file doesn't exist yet.
        # remove this once v.07 is released
        v7: Annotated[
            bool,
            typer.Option(
                "--seven",
                "-s",
                help="Emulate v0.7",
                hidden=(Flags.v07.is_enabled() or not CDFToml.load().is_loaded_from_file),
            ),
        ] = False,
    ) -> None:
        """Getting started checklist"""
        cmd = InitCommand()
        # Tracking  command with the usual lambda run construct
        # is intentionally left out because we don't want to expose the user to the warning
        # before they've had the chance to opt in (which is something they'll do later using this command).
        cmd.execute(dry_run=dry_run, emulate_dot_seven=v7)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Generic, cast
from uuid import uuid4

from cognite.client.data_classes._base import (
    T_CogniteResource,
)
from cognite.client.data_classes.data_modeling import (
    EdgeApply,
    InstanceApply,
    NodeApply,
    NodeId,
    View,
    ViewId,
)

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.charts import Chart, ChartWrite
from cognite_toolkit._cdf_tk.client.data_classes.charts_data import (
    ChartCoreTimeseries,
    ChartSource,
    ChartTimeseries,
)
from cognite_toolkit._cdf_tk.client.data_classes.migration import ResourceViewMappingApply
from cognite_toolkit._cdf_tk.commands._migrate.conversion import DirectRelationCache, asset_centric_to_dm
from cognite_toolkit._cdf_tk.commands._migrate.data_classes import AssetCentricMapping
from cognite_toolkit._cdf_tk.commands._migrate.default_mappings import create_default_mappings
from cognite_toolkit._cdf_tk.commands._migrate.issues import ChartMigrationIssue, ConversionIssue, MigrationIssue
from cognite_toolkit._cdf_tk.commands._migrate.selectors import AssetCentricMigrationSelector
from cognite_toolkit._cdf_tk.constants import MISSING_INSTANCE_SPACE
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.storageio._base import T_Selector, T_WriteCogniteResource
from cognite_toolkit._cdf_tk.storageio.selectors import ChartSelector
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.useful_types import (
    T_AssetCentricResourceExtended,
)


class DataMapper(Generic[T_Selector, T_CogniteResource, T_WriteCogniteResource], ABC):
    def prepare(self, source_selector: T_Selector) -> None:
        """Prepare the data mapper with the given source selector.

        Args:
            source_selector: The selector for the source data.

        """
        # Override in subclass if needed.
        pass

    @abstractmethod
    def map(
        self, source: Sequence[T_CogniteResource]
    ) -> Sequence[tuple[T_WriteCogniteResource | None, MigrationIssue]]:
        """Map a chunk of source data to the target format.

        Args:
            source: The source data chunk to be mapped.

        Returns:
            A tuple containing the mapped data and a list of any issues encountered during mapping.

        """
        raise NotImplementedError("Subclasses must implement this method.")


class AssetCentricMapper(
    DataMapper[AssetCentricMigrationSelector, AssetCentricMapping[T_AssetCentricResourceExtended], InstanceApply]
):
    def __init__(self, client: ToolkitClient) -> None:
        self.client = client
        self._ingestion_view_by_id: dict[ViewId, View] = {}
        self._view_mapping_by_id: dict[str, ResourceViewMappingApply] = {}
        self._direct_relation_cache = DirectRelationCache(client)

    def prepare(self, source_selector: AssetCentricMigrationSelector) -> None:
        ingestion_view_ids = source_selector.get_ingestion_mappings()
        ingestion_views = self.client.migration.resource_view_mapping.retrieve(ingestion_view_ids)
        defaults = {mapping.external_id: mapping for mapping in create_default_mappings()}
        # Custom mappings from CDF override the default mappings
        self._view_mapping_by_id = defaults | {view.external_id: view.as_write() for view in ingestion_views}
        missing_mappings = set(ingestion_view_ids) - set(self._view_mapping_by_id.keys())
        if missing_mappings:
            raise ToolkitValueError(
                f"The following ingestion views were not found: {humanize_collection(missing_mappings)}"
            )

        view_ids = list({mapping.view_id for mapping in self._view_mapping_by_id.values()})
        views = self.client.data_modeling.views.retrieve(view_ids)
        self._ingestion_view_by_id = {view.as_id(): view for view in views}
        missing_views = set(view_ids) - set(self._ingestion_view_by_id.keys())
        if missing_views:
            raise ToolkitValueError(
                f"The following ingestion views were not found in Data Modeling: {humanize_collection(missing_views)}"
            )

    def map(
        self, source: Sequence[AssetCentricMapping[T_AssetCentricResourceExtended]]
    ) -> Sequence[tuple[InstanceApply | None, ConversionIssue]]:
        """Map a chunk of asset-centric data to InstanceApplyList format."""
        # We update the direct relation cache in bulk for all resources in the chunk.
        self._direct_relation_cache.update(item.resource for item in source)
        output: list[tuple[InstanceApply | None, ConversionIssue]] = []
        for item in source:
            instance, conversion_issue = self._map_single_item(item)
            output.append((instance, conversion_issue))
        return output

    def _map_single_item(
        self, item: AssetCentricMapping[T_AssetCentricResourceExtended]
    ) -> tuple[NodeApply | EdgeApply | None, ConversionIssue]:
        mapping = item.mapping
        ingestion_view = mapping.get_ingestion_view()
        try:
            view_source = self._view_mapping_by_id[ingestion_view]
            view_properties = self._ingestion_view_by_id[view_source.view_id].properties
        except KeyError as e:
            raise RuntimeError(
                f"Failed to lookup mapping or view for ingestion view '{ingestion_view}'. Did you forget to call .prepare()?"
            ) from e
        instance, conversion_issue = asset_centric_to_dm(
            item.resource,
            instance_id=mapping.instance_id,
            view_source=view_source,
            view_properties=view_properties,
            direct_relation_cache=self._direct_relation_cache,
        )
        if mapping.instance_id.space == MISSING_INSTANCE_SPACE:
            conversion_issue.missing_instance_space = f"Missing instance space for dataset ID {mapping.data_set_id!r}"
        return instance, conversion_issue


class ChartMapper(DataMapper[ChartSelector, Chart, ChartWrite]):
    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    def map(self, source: Sequence[Chart]) -> Sequence[tuple[ChartWrite | None, MigrationIssue]]:
        self._populate_cache(source)
        output: list[tuple[ChartWrite | None, MigrationIssue]] = []
        for item in source:
            mapped_item, issue = self._map_single_item(item)
            output.append((mapped_item, issue))
        return output

    def _populate_cache(self, source: Sequence[Chart]) -> None:
        """Populate the internal cache with timeseries from the source charts.

        Note that the consumption views are also cached as part of the timeseries lookup.
        """
        timeseries_ids: set[int] = set()
        timeseries_external_ids: set[str] = set()
        for chart in source:
            for item in chart.data.time_series_collection or []:
                if item.ts_id:
                    timeseries_ids.add(item.ts_id)
                if item.ts_external_id:
                    timeseries_external_ids.add(item.ts_external_id)
        if timeseries_ids:
            self.client.migration.lookup.time_series(list(timeseries_ids))
        if timeseries_external_ids:
            self.client.migration.lookup.time_series(external_id=list(timeseries_external_ids))

    def _map_single_item(self, item: Chart) -> tuple[ChartWrite | None, ChartMigrationIssue]:
        issue = ChartMigrationIssue(chart_external_id=item.external_id)
        time_series_collection = item.data.time_series_collection or []
        timeseries_core_collection = self._create_timeseries_core_collection(time_series_collection, issue)
        if issue.has_issues:
            return None, issue

        updated_source_collection = self._update_source_collection(
            item.data.source_collection or [], time_series_collection, timeseries_core_collection
        )

        mapped_chart = item.as_write()
        mapped_chart.data.core_timeseries_collection = timeseries_core_collection
        mapped_chart.data.time_series_collection = None
        mapped_chart.data.source_collection = updated_source_collection
        return mapped_chart, issue

    def _create_timeseries_core_collection(
        self, time_series_collection: list[ChartTimeseries], issue: ChartMigrationIssue
    ) -> list[ChartCoreTimeseries]:
        timeseries_core_collection: list[ChartCoreTimeseries] = []
        for ts_item in time_series_collection or []:
            node_id, consumer_view_id = self._get_node_id_consumer_view_id(ts_item)

            if node_id is None:
                if ts_item.ts_id is not None:
                    issue.missing_timeseries_ids.append(ts_item.ts_id)
                elif ts_item.ts_external_id is not None:
                    issue.missing_timeseries_external_ids.append(ts_item.ts_external_id)
                else:
                    issue.missing_timeseries_identifier.append(ts_item.id or "unknown")
                continue

            core_timeseries = self._create_new_timeseries_core(ts_item, node_id, consumer_view_id)
            timeseries_core_collection.append(core_timeseries)
        return timeseries_core_collection

    def _create_new_timeseries_core(
        self, ts_item: ChartTimeseries, node_id: NodeId, consumer_view_id: ViewId | None
    ) -> ChartCoreTimeseries:
        dumped = ts_item.dump(camel_case=True)
        for asset_centric_key in ["tsId", "tsExternalId", "originalUnit"]:
            dumped.pop(asset_centric_key, None)

        dumped["nodeReference"] = node_id
        dumped["viewReference"] = consumer_view_id
        new_uuid = str(uuid4())
        dumped["id"] = new_uuid
        dumped["type"] = "coreTimeseries"
        core_timeseries = ChartCoreTimeseries._load(dumped)
        return core_timeseries

    def _get_node_id_consumer_view_id(self, ts_item: ChartTimeseries) -> tuple[NodeId | None, ViewId | None]:
        """Look up the node ID and consumer view ID for a given timeseries item.

        Prioritizes lookup by internal ID, then by external ID.

        Args:
            ts_item: The ChartTimeseries item to look up.

        Returns:
            A tuple containing the consumer view ID and node ID, or None if not found.
        """
        node_id: NodeId | None = None
        consumer_view_id: ViewId | None = None
        for id_name, id_value in [("id", ts_item.ts_id), ("external_id", ts_item.ts_external_id)]:
            if id_value is None:
                continue
            arg = {id_name: id_value}
            node_id = self.client.migration.lookup.time_series(**arg)  # type: ignore[arg-type]
            consumer_view_id = self.client.migration.lookup.time_series.consumer_view(**arg)  # type: ignore[arg-type]
            if node_id is not None:
                break
        return node_id, consumer_view_id

    def _update_source_collection(
        self,
        source_collection: list[ChartSource],
        time_series_collection: list[ChartTimeseries],
        timeseries_core_collection: list[ChartCoreTimeseries],
    ) -> list[ChartSource]:
        remove_ids = {ts_item.id for ts_item in time_series_collection if ts_item.id is not None}
        updated_source_collection = [ts_item for ts_item in source_collection if ts_item.id not in remove_ids]
        for core_ts_item in timeseries_core_collection:
            # We cast there two as we set them in the _create_timeseries_core_collection method
            new_source_item = ChartSource(id=cast(str, core_ts_item.id), type=cast(str, core_ts_item.type))
            updated_source_collection.append(new_source_item)
        return updated_source_collection



import json
import uuid
from collections.abc import Iterator
from pathlib import Path
from unittest.mock import MagicMock, patch

import httpx
import pytest
import responses
import respx
from cognite.client.data_classes import Annotation, AnnotationList, Asset, AssetList
from cognite.client.data_classes.data_modeling import (
    DataModel,
    DataModelList,
    EdgeApply,
    NodeApply,
    NodeOrEdgeData,
    View,
    ViewId,
)
from cognite.client.data_classes.data_modeling.statistics import InstanceStatistics, ProjectStatistics

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.charts import Chart
from cognite_toolkit._cdf_tk.client.data_classes.charts_data import ChartData, ChartSource, ChartTimeseries
from cognite_toolkit._cdf_tk.client.data_classes.migration import InstanceSource
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands._migrate.command import MigrationCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import AssetCentricMapper, ChartMapper
from cognite_toolkit._cdf_tk.commands._migrate.data_model import (
    COGNITE_MIGRATION_MODEL,
    INSTANCE_SOURCE_VIEW_ID,
    MODEL_ID,
    RESOURCE_VIEW_MAPPING_VIEW_ID,
)
from cognite_toolkit._cdf_tk.commands._migrate.default_mappings import (
    ASSET_ANNOTATIONS_ID,
    ASSET_ID,
    FILE_ANNOTATIONS_ID,
    create_default_mappings,
)
from cognite_toolkit._cdf_tk.commands._migrate.migration_io import AnnotationMigrationIO, AssetCentricMigrationIO
from cognite_toolkit._cdf_tk.commands._migrate.selectors import MigrationCSVFileSelector
from cognite_toolkit._cdf_tk.exceptions import ToolkitMigrationError, ToolkitValueError
from cognite_toolkit._cdf_tk.storageio import ChartIO
from cognite_toolkit._cdf_tk.storageio.selectors import ChartExternalIdSelector
from cognite_toolkit._cdf_tk.utils.fileio import CSVReader


@pytest.fixture
def cognite_migration_model(
    toolkit_config: ToolkitClientConfig,
    rsps: responses.RequestsMock,
    cognite_core_no_3D: DataModel[View],
    cognite_extractor_views: list[View],
) -> Iterator[responses.RequestsMock]:
    """Mock the Cognite Migration Model in the CDF project."""
    config = toolkit_config
    # Migration model
    migration_model = COGNITE_MIGRATION_MODEL.dump()
    migration_model["createdTime"] = 1
    migration_model["lastUpdatedTime"] = 1
    migration_model["isGlobal"] = False
    rsps.post(config.create_api_url("models/datamodels/byids"), json={"items": migration_model})
    yield rsps


@pytest.fixture
def resource_view_mappings(
    toolkit_config: ToolkitClientConfig,
    cognite_migration_model: responses.RequestsMock,
    cognite_core_no_3D: DataModel[View],
    cognite_extractor_views: list[View],
) -> Iterator[responses.RequestsMock]:
    """Mock all the default Resource View Mappings in the Cognite Migration Model."""
    rsps = cognite_migration_model
    config = toolkit_config
    mapping_by_id = {mapping.external_id: mapping for mapping in create_default_mappings()}
    node_items: list[dict] = []
    for mapping in mapping_by_id.values():
        # Lookup of the mapping in the Migration Model
        mapping_node_response = mapping.dump(context="api")
        mapping_node_response.update({"createdTime": 0, "lastUpdatedTime": 0, "version": 1})
        sources = mapping_node_response.pop("sources", [])
        if sources:
            mapping_view_id = mapping.sources[0].source
            mapping_node_response["properties"] = {
                mapping_view_id.space: {
                    f"{mapping_view_id.external_id}/{mapping_view_id.version}": sources[0]["properties"]
                }
            }
        node_items.append(mapping_node_response)
    rsps.post(
        config.create_api_url("models/instances/byids"),
        json={"items": node_items},
        status=200,
    )
    rsps.post(
        config.create_api_url("models/views/byids"),
        json={
            "items": [view.dump() for view in cognite_core_no_3D.views]
            + [view.dump() for view in cognite_extractor_views]
        },
    )
    yield rsps


@pytest.fixture
def mock_statistics(
    toolkit_config: ToolkitClientConfig,
    rsps: responses.RequestsMock,
) -> Iterator[responses.RequestsMock]:
    config = toolkit_config
    stats_response = {
        "spaces": {
            "count": 0,
            "limit": 1_000,
        },
        "containers": {
            "count": 0,
            "limit": 10_000,
        },
        "views": {
            "count": 0,
            "limit": 100_000,
        },
        "dataModels": {
            "count": 1,
            "limit": 10_000,
        },
        "containerProperties": {
            "count": 0,
            "limit": 1_000_000,
        },
        "instances": {
            "nodes": 1000,
            "edges": 0,
            "softDeletedNodes": 0,
            "softDeletedEdges": 0,
            "instancesLimit": 5_000_000,
            "softDeletedInstancesLimit": 100_000_000,
            "instances": 1000,
            "softDeletedInstances": 0,
        },
        "concurrentReadLimit": 50,
        "concurrentWriteLimit": 20,
        "concurrentDeleteLimit": 10,
    }
    rsps.get(
        config.create_api_url("/models/statistics"),
        json=stats_response,
        status=200,
    )
    yield rsps


@pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
class TestMigrationCommand:
    @pytest.mark.usefixtures("mock_statistics")
    def test_migrate_assets(
        self,
        toolkit_config: ToolkitClientConfig,
        resource_view_mappings: responses.RequestsMock,
        tmp_path: Path,
        respx_mock: respx.MockRouter,
    ) -> None:
        rsps = resource_view_mappings
        config = toolkit_config
        assets = AssetList(
            [
                Asset(
                    id=1000 + i,
                    external_id=f"asset_{i}",
                    name=f"Asset {i}",
                    description=f"This is Asset {i}",
                    last_updated_time=1,
                    created_time=0,
                    parent_external_id="asset_0" if i > 0 else None,
                )
                for i in range(2)
            ]
        )
        space = "my_space"
        csv_content = "id,space,externalId,ingestionView\n" + "\n".join(
            f"{1000 + i},{space},asset_{i},{ASSET_ID}" for i in range(len(assets))
        )

        # Asset retrieve ids
        rsps.post(
            config.create_api_url("/assets/byids"),
            json={"items": [asset.dump() for asset in assets]},
            status=200,
        )
        # Instance creation
        respx.post(
            config.create_api_url("/models/instances"),
        ).mock(
            return_value=httpx.Response(
                status_code=200,
                json={
                    "items": [
                        {
                            "instanceType": "node",
                            "space": space,
                            "externalId": f"asset_{i}",
                            "version": 1,
                            "wasModified": True,
                            "createdTime": 1,
                            "lastUpdatedTime": 1,
                        }
                        for i in range(len(assets))
                    ]
                },
            )
        )
        csv_file = tmp_path / "migration.csv"
        csv_file.write_text(csv_content, encoding="utf-8")

        client = ToolkitClient(config)
        command = MigrationCommand(silent=True)

        result = command.migrate(
            selected=MigrationCSVFileSelector(datafile=csv_file, kind="Assets"),
            data=AssetCentricMigrationIO(client),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path / "logs",
            dry_run=False,
            verbose=False,
        )

        # Check that the assets were uploaded
        last_call = respx_mock.calls[-1]
        assert last_call.request.url == config.create_api_url("/models/instances")
        assert last_call.request.method == "POST"
        actual_instances = json.loads(last_call.request.content)["items"]
        expected_instance = [
            NodeApply(
                space=space,
                external_id=asset.external_id,
                sources=[
                    NodeOrEdgeData(
                        source=ViewId("cdf_cdm", "CogniteAsset", "v1"),
                        properties={
                            "name": asset.name,
                            "description": asset.description,
                        },
                    ),
                    NodeOrEdgeData(
                        source=INSTANCE_SOURCE_VIEW_ID,
                        properties={
                            "id": asset.id,
                            "resourceType": "asset",
                            "dataSetId": None,
                            "classicExternalId": asset.external_id,
                        },
                    ),
                ],
            ).dump()
            for asset in assets
        ]
        assert actual_instances == expected_instance
        actual_results = [result.get_progress(f"asset_{asset.id}") for asset in assets]
        expected_results = [{"download": "success", "convert": "success", "upload": "success"} for _ in assets]
        assert actual_results == expected_results
        csv_file = next((tmp_path / "logs").glob("*.csv"), None)
        assert csv_file is not None, "Expected a CSV log file to be created"
        csv_results = list(CSVReader(csv_file).read_chunks_unprocessed())
        assert csv_results == [
            {"ID": f"asset_{asset.id}", "download": "success", "convert": "success", "upload": "success"}
            for asset in assets
        ]

    @pytest.mark.usefixtures("mock_statistics")
    def test_migrate_annotations(
        self,
        toolkit_config: ToolkitClientConfig,
        resource_view_mappings: responses.RequestsMock,
        tmp_path: Path,
        respx_mock: respx.MockRouter,
    ) -> None:
        rsps = resource_view_mappings
        config = toolkit_config
        asset_annotation = Annotation(
            id=2000,
            annotated_resource_type="file",
            annotated_resource_id=3000,
            data={
                "assetRef": {"id": 4000},
                "textRegion": {"xMin": 10.0, "xMax": 100.0, "yMin": 20.0, "yMax": 200.0},
            },
            status="Approved",
            creating_user="doctrino",
            creating_app="my_app",
            creating_app_version="v1",
            annotation_type="diagrams.AssetLink",
        )
        file_annotation = Annotation(
            id=2001,
            annotated_resource_type="file",
            annotated_resource_id=3001,
            data={
                "fileRef": {"id": 5000},
                "textRegion": {"xMin": 15.0, "xMax": 150.0, "yMin": 25.0, "yMax": 250.0},
            },
            status="Approved",
            creating_user="doctrino",
            creating_app="my_app",
            creating_app_version="v1",
            annotation_type="diagrams.FileLink",
        )
        annotations = AnnotationList([asset_annotation, file_annotation])
        space = "my_space"
        csv_content = "id,space,externalId,ingestionView\n" + "\n".join(
            (
                f"{2000},{space},annotation_{2000},{ASSET_ANNOTATIONS_ID}",
                f"{2001},{space},annotation_{2001},{FILE_ANNOTATIONS_ID}",
            )
        )
        # Annotation retrieve ids
        rsps.post(
            config.create_api_url("/annotations/byids"),
            json={"items": [annotation.dump() for annotation in annotations]},
            status=200,
        )
        # Lookup asset and file instance ID
        for items in [
            [("asset", 4000)],
            [("file", 5000), ("file", 3000), ("file", 3001)],
        ]:
            rsps.post(
                config.create_api_url("/models/instances/query"),
                json={
                    "items": {
                        "instanceSource": [
                            {
                                "instanceType": "node",
                                "space": space,
                                "externalId": f"{resource_type}_{resource_id}",
                                "version": 0,
                                "createdTime": 0,
                                "lastUpdatedTime": 0,
                                "properties": {
                                    "cognite_migration": {
                                        "InstanceSource/v1": {
                                            "id": resource_id,
                                            "resourceType": resource_type,
                                        }
                                    },
                                },
                            }
                            for (resource_type, resource_id) in items
                        ],
                    },
                    "nextCursor": {"instanceSource": None},
                },
                status=200,
            )

        # Instance creation
        respx.post(
            config.create_api_url("/models/instances"),
        ).mock(
            return_value=httpx.Response(
                status_code=200,
                json={
                    "items": [
                        {
                            "instanceType": "edge",
                            "space": space,
                            "externalId": f"annotation_{2000 + i}",
                            "version": 1,
                            "wasModified": True,
                            "createdTime": 1,
                            "lastUpdatedTime": 1,
                        }
                        for i in range(len(annotations))
                    ]
                },
            )
        )
        csv_file = tmp_path / "migration.csv"
        csv_file.write_text(csv_content, encoding="utf-8")

        client = ToolkitClient(config)
        command = MigrationCommand(silent=True)

        result = command.migrate(
            selected=MigrationCSVFileSelector(datafile=csv_file, kind="Annotations"),
            data=AnnotationMigrationIO(client),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path / "logs",
            dry_run=False,
            verbose=True,
        )
        actual_results = [result.get_progress(f"Annotation_{annotation.id}") for annotation in annotations]
        expected_results = [{"download": "success", "convert": "success", "upload": "success"} for _ in annotations]
        assert actual_results == expected_results

        # Check that the annotations were uploaded
        last_call = respx_mock.calls[-1]
        assert last_call.request.url == config.create_api_url("/models/instances")
        assert last_call.request.method == "POST"
        actual_instances = json.loads(last_call.request.content)["items"]
        expected_instance = [
            EdgeApply(
                space=space,
                external_id=f"annotation_{asset_annotation.id}",
                start_node=(space, f"file_{asset_annotation.annotated_resource_id}"),
                end_node=(space, f"asset_{asset_annotation.data['assetRef']['id']}"),
                type=(space, asset_annotation.annotation_type),
                sources=[
                    NodeOrEdgeData(
                        source=ViewId("cdf_cdm", "CogniteDiagramAnnotation", "v1"),
                        properties={
                            "sourceContext": asset_annotation.creating_app_version,
                            "sourceCreatedUser": asset_annotation.creating_user,
                            "sourceId": asset_annotation.creating_app,
                            "status": asset_annotation.status,
                            "startNodeXMax": asset_annotation.data["textRegion"]["xMax"],
                            "startNodeXMin": asset_annotation.data["textRegion"]["xMin"],
                            "startNodeYMax": asset_annotation.data["textRegion"]["yMax"],
                            "startNodeYMin": asset_annotation.data["textRegion"]["yMin"],
                        },
                    ),
                ],
            ).dump(),
            EdgeApply(
                space=space,
                external_id=f"annotation_{file_annotation.id}",
                start_node=(space, f"file_{file_annotation.annotated_resource_id}"),
                end_node=(space, f"file_{file_annotation.data['fileRef']['id']}"),
                type=(space, file_annotation.annotation_type),
                sources=[
                    NodeOrEdgeData(
                        source=ViewId("cdf_cdm", "CogniteDiagramAnnotation", "v1"),
                        properties={
                            "sourceContext": file_annotation.creating_app_version,
                            "sourceCreatedUser": file_annotation.creating_user,
                            "sourceId": file_annotation.creating_app,
                            "status": file_annotation.status,
                            "startNodeXMax": file_annotation.data["textRegion"]["xMax"],
                            "startNodeXMin": file_annotation.data["textRegion"]["xMin"],
                            "startNodeYMax": file_annotation.data["textRegion"]["yMax"],
                            "startNodeYMin": file_annotation.data["textRegion"]["yMin"],
                        },
                    ),
                ],
            ).dump(),
        ]
        assert actual_instances == expected_instance

    def test_migrate_charts(
        self,
        toolkit_config: ToolkitClientConfig,
        cognite_migration_model: responses.RequestsMock,
        tmp_path: Path,
        respx_mock: respx.MockRouter,
    ) -> None:
        rsps = cognite_migration_model
        config = toolkit_config
        charts = [
            Chart(
                external_id="my_chart",
                created_time=1,
                last_updated_time=1,
                visibility="PUBLIC",
                data=ChartData(
                    name="My Chart",
                    time_series_collection=[
                        ChartTimeseries(
                            ts_external_id="ts_1", type="timeseries", id="87654321-4321-8765-4321-876543218765"
                        ),
                        ChartTimeseries(ts_id=1, type="timeseries", id="12345678-1234-5678-1234-567812345678"),
                    ],
                    source_collection=[
                        ChartSource(type="timeseries", id="87654321-4321-8765-4321-876543218765"),
                        ChartSource(type="timeseries", id="12345678-1234-5678-1234-567812345678"),
                    ],
                ),
                owner_id="1234",
            )
        ]
        # Chart retrieve ids
        rsps.add(
            responses.POST,
            config.create_app_url("/storage/charts/charts/list"),
            json={
                "items": [chart.dump() for chart in charts],
            },
            status=200,
        )
        # TimeSeries Instance ID lookup
        rsps.add(
            method=responses.POST,
            url=config.create_api_url("models/instances/query"),
            json={
                "items": {
                    "instanceSource": [
                        InstanceSource(
                            space="my_space",
                            external_id="node_123",
                            version=1,
                            last_updated_time=1,
                            created_time=1,
                            resource_type="timeseries",
                            id_=1,
                            classic_external_id=None,
                            preferred_consumer_view_id=ViewId("cdf_cdm", "CogniteTimeSeries", "v1"),
                        ).dump(),
                        InstanceSource(
                            space="my_space",
                            external_id="node_ts_1",
                            version=1,
                            last_updated_time=1,
                            created_time=1,
                            resource_type="timeseries",
                            id_=2,
                            classic_external_id="ts_1",
                            preferred_consumer_view_id=ViewId("my_schema_space", "MyTimeSeries", "v1"),
                        ).dump(),
                    ]
                },
                "nextCursor": {"instanceSource": None},
            },
            status=200,
        )

        # Chart upsert
        respx.put(
            config.create_app_url("/storage/charts/charts"),
        ).mock(return_value=httpx.Response(status_code=200, json={"items": [chart.dump() for chart in charts]}))

        client = ToolkitClient(config)
        command = MigrationCommand(silent=True)
        new_uuids = [
            uuid.UUID("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"),
            uuid.UUID("bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"),
        ]
        with patch(f"{ChartMapper.__module__}.uuid4", side_effect=new_uuids):
            result = command.migrate(
                selected=ChartExternalIdSelector(external_ids=("my_chart",)),
                data=ChartIO(client),
                mapper=ChartMapper(client),
                log_dir=tmp_path / "logs",
                dry_run=False,
                verbose=True,
            )
        actual = result.get_progress("my_chart")
        expected = {"download": "success", "convert": "success", "upload": "success"}
        assert actual == expected

        calls = respx_mock.calls
        assert len(calls) == 1
        last_call = calls[-1]
        assert last_call.request.url == config.create_app_url("/storage/charts/charts")
        assert last_call.request.method == "PUT"
        actual_charts = json.loads(last_call.request.content)["items"]
        expected_charts = [
            {
                "externalId": "my_chart",
                "visibility": "PUBLIC",
                "data": {
                    "name": "My Chart",
                    "coreTimeseriesCollection": [
                        {
                            "type": "coreTimeseries",
                            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                            "nodeReference": {"space": "my_space", "externalId": "node_ts_1"},
                            "viewReference": {
                                "space": "my_schema_space",
                                "externalId": "MyTimeSeries",
                                "version": "v1",
                            },
                        },
                        {
                            "type": "coreTimeseries",
                            "id": "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb",
                            "nodeReference": {"space": "my_space", "externalId": "node_123"},
                            "viewReference": {"space": "cdf_cdm", "externalId": "CogniteTimeSeries", "version": "v1"},
                        },
                    ],
                    "sourceCollection": [
                        {
                            "type": "coreTimeseries",
                            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                        },
                        {
                            "type": "coreTimeseries",
                            "id": "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb",
                        },
                    ],
                },
            }
        ]
        assert actual_charts == expected_charts

    def test_validate_migration_model_available(self) -> None:
        with monkeypatch_toolkit_client() as client:
            client.data_modeling.data_models.retrieve.return_value = DataModelList([])
            with pytest.raises(ToolkitMigrationError):
                MigrationCommand.validate_migration_model_available(client)

    def test_validate_migration_model_available_multiple_models(self) -> None:
        """Test that multiple models raises an error."""
        with monkeypatch_toolkit_client() as client:
            # Create mock models with the expected MODEL_ID
            model1 = MagicMock(spec=DataModel)
            model1.as_id.return_value = MODEL_ID
            model2 = MagicMock(spec=DataModel)
            model2.as_id.return_value = MODEL_ID

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model1, model2])

            with pytest.raises(ToolkitMigrationError) as exc_info:
                MigrationCommand.validate_migration_model_available(client)

            assert "Multiple migration models" in str(exc_info.value)

    def test_validate_migration_model_available_missing_views(self) -> None:
        """Test that a model with missing views raises an error."""
        with monkeypatch_toolkit_client() as client:
            model = MagicMock(spec=DataModel)
            model.as_id.return_value = MODEL_ID
            # Model has views but missing the required ones
            model.views = [INSTANCE_SOURCE_VIEW_ID]  # Missing VIEW_SOURCE_VIEW_ID

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model])

            with pytest.raises(ToolkitMigrationError, match=r"Invalid migration model. Missing views"):
                MigrationCommand.validate_migration_model_available(client)

    def test_validate_migration_model_available_success(self) -> None:
        """Test that a valid model with all required views succeeds."""
        with monkeypatch_toolkit_client() as client:
            # Mocking the migration Model to get a response format of the model.
            # An alternative would be to write a conversion of write -> read format of the model
            # which is a significant amount of logic.
            model = MagicMock(spec=DataModel)
            model.as_id.return_value = MODEL_ID
            # Model has all required views
            model.views = [INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID]

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model])

            # Should not raise any exception
            MigrationCommand.validate_migration_model_available(client)

            client.data_modeling.data_models.retrieve.assert_called_once_with([MODEL_ID], inline_views=False)

    def test_validate_available_capacity_missing_capacity(self) -> None:
        cmd = MigrationCommand(silent=True)

        with monkeypatch_toolkit_client() as client:
            stats = MagicMock(spec=ProjectStatistics)
            stats.instances = InstanceStatistics(
                nodes=1000,
                edges=0,
                soft_deleted_edges=0,
                soft_deleted_nodes=0,
                instances_limit=1500,
                soft_deleted_instances_limit=10_000,
                instances=1000,
                soft_deleted_instances=0,
            )
            client.data_modeling.statistics.project.return_value = stats
            with pytest.raises(ToolkitValueError) as exc_info:
                cmd.validate_available_capacity(client, 10_000)

        assert "Cannot proceed with migration" in str(exc_info.value)

    def test_validate_available_capacity_sufficient_capacity(self) -> None:
        cmd = MigrationCommand(silent=True)

        with monkeypatch_toolkit_client() as client:
            stats = MagicMock(spec=ProjectStatistics)
            stats.instances = InstanceStatistics(
                nodes=1000,
                edges=0,
                soft_deleted_edges=0,
                soft_deleted_nodes=0,
                instances_limit=5_000_000,
                soft_deleted_instances_limit=100_000_000,
                instances=1000,
                soft_deleted_instances=0,
            )
            client.data_modeling.statistics.project.return_value = stats
            cmd.validate_available_capacity(client, 10_000)



from typing import Any, ClassVar

import pytest
import responses
from cognite.client.data_classes.data_modeling import NodeId, ViewId

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.migration import InstanceSource


@pytest.fixture()
def lookup_client(
    toolkit_config: ToolkitClientConfig,
    rsps: responses.RequestsMock,
) -> tuple[ToolkitClient, responses.RequestsMock]:
    config = toolkit_config
    rsps.add(
        method=responses.POST,
        url=config.create_api_url("models/instances/query"),
        json=TestMigrationLookup.QUERY_RESPONSE,
        status=200,
    )

    return ToolkitClient(config=config), rsps


@pytest.mark.usefixtures("disable_pypi_check")
class TestMigrationLookup:
    SPACE = "my_space"
    EXISTING_ID = 123
    EXISTING_EXTERNAL_ID = "node_123"
    EXISTING_NODE_ID = NodeId(SPACE, "node_123")
    CONSUMER_VIEW = ViewId("cdf_cdm", "CogniteAsset", "v1")
    QUERY_RESPONSE: ClassVar[dict[str, Any]] = {
        "items": {
            "instanceSource": [
                InstanceSource(
                    space=EXISTING_NODE_ID.space,
                    external_id=EXISTING_NODE_ID.external_id,
                    version=1,
                    last_updated_time=1,
                    created_time=1,
                    resource_type="asset",
                    id_=EXISTING_ID,
                    classic_external_id=EXISTING_EXTERNAL_ID,
                    preferred_consumer_view_id=CONSUMER_VIEW,
                ).dump()
            ]
        },
        "nextCursor": {"instanceSource": None},
    }

    @pytest.mark.parametrize(
        "args, expected_return",
        [
            pytest.param({"id": EXISTING_ID}, EXISTING_NODE_ID, id="Exiting single ID"),
            pytest.param({"external_id": EXISTING_EXTERNAL_ID}, EXISTING_NODE_ID, id="Existing single external ID"),
            pytest.param({"id": -1}, None, id="Non-existing single ID"),
            pytest.param({"external_id": "non_existing_external_id"}, None, id="Non-existing single external ID"),
            pytest.param(
                {"id": [EXISTING_ID, -1]}, {EXISTING_ID: EXISTING_NODE_ID}, id="Mixed existing and non-existing IDs"
            ),
            pytest.param(
                {"external_id": [EXISTING_EXTERNAL_ID, "non_existing_external_id"]},
                {EXISTING_EXTERNAL_ID: EXISTING_NODE_ID},
                id="Mixed existing and non-existing external IDs",
            ),
        ],
    )
    def test_return_type_given_input(
        self,
        args: dict[str, Any],
        expected_return: dict | NodeId | None,
        lookup_client: tuple[ToolkitClient, responses.RequestsMock],
    ) -> None:
        client, _ = lookup_client
        actual_return = client.migration.lookup.assets(**args)
        assert actual_return == expected_return

    def test_multi_lookup_single_api_call(self, lookup_client: tuple[ToolkitClient, responses.RequestsMock]) -> None:
        client, rsps = lookup_client
        _ = client.migration.lookup.assets(self.EXISTING_ID)
        _ = client.migration.lookup.assets(self.EXISTING_ID)

        assert len(rsps.calls) == 1, "Expected only one API call for multiple lookups of the same ID"

    def test_single_api_call_non_existing(self, lookup_client: tuple[ToolkitClient, responses.RequestsMock]) -> None:
        client, rsps = lookup_client
        _ = client.migration.lookup.assets(-1)
        _ = client.migration.lookup.assets(-1)

        assert len(rsps.calls) == 1, "Expected only one API call for multiple lookups of the same non-existing ID"

    def test_invalid_input(self, toolkit_config: ToolkitClientConfig):
        client = ToolkitClient(config=toolkit_config)
        with pytest.raises(TypeError):
            _ = client.migration.lookup.assets(id=self.EXISTING_ID, external_id=self.EXISTING_EXTERNAL_ID)

    @pytest.mark.parametrize(
        "args, expected_return",
        [
            pytest.param({"id": EXISTING_ID}, CONSUMER_VIEW, id="Existing single ID"),
            pytest.param({"external_id": EXISTING_EXTERNAL_ID}, CONSUMER_VIEW, id="Existing single external ID"),
            pytest.param({"id": -1}, None, id="Non-existing single ID"),
            pytest.param({"external_id": "non_existing_external_id"}, None, id="Non-existing single external ID"),
            pytest.param(
                {"id": {EXISTING_ID, -1}}, {EXISTING_ID: CONSUMER_VIEW}, id="Mixed existing and non-existing IDs"
            ),
            pytest.param(
                {"external_id": [EXISTING_EXTERNAL_ID, "non_existing_external_id"]},
                {EXISTING_EXTERNAL_ID: CONSUMER_VIEW},
                id="Mixed existing and non-existing external IDs",
            ),
        ],
    )
    def test_get_preferred_consumer_view(
        self,
        args: dict[str, Any],
        expected_return: ViewId,
        lookup_client: tuple[ToolkitClient, responses.RequestsMock],
    ) -> None:
        client, _ = lookup_client
        actual_return = client.migration.lookup.assets.consumer_view(**args)
        assert actual_return == expected_return

    def test_consumer_view_single_api_call(self, lookup_client: tuple[ToolkitClient, responses.RequestsMock]) -> None:
        client, rsps = lookup_client
        _ = client.migration.lookup.assets.consumer_view(self.EXISTING_ID)
        _ = client.migration.lookup.assets.consumer_view(self.EXISTING_ID)

        assert len(rsps.calls) == 1, "Expected only one API call for multiple lookups of the same ID"

    def test_consumer_view_non_existing_single_api_call(
        self, lookup_client: tuple[ToolkitClient, responses.RequestsMock]
    ) -> None:
        client, rsps = lookup_client
        _ = client.migration.lookup.assets.consumer_view(-1)
        _ = client.migration.lookup.assets.consumer_view(-1)

        assert len(rsps.calls) == 1, "Expected only one API call for multiple lookups of the same non-existing ID"

    def test_consumer_view_invalid_input(self, toolkit_config: ToolkitClientConfig):
        client = ToolkitClient(config=toolkit_config)
        with pytest.raises(TypeError):
            _ = client.migration.lookup.assets.consumer_view(id=self.EXISTING_ID, external_id=self.EXISTING_EXTERNAL_ID)



from collections.abc import Iterable, Mapping, Set
from typing import Any, ClassVar, cast

from cognite.client.data_classes import Annotation, Asset, Event, FileMetadata, TimeSeries
from cognite.client.data_classes.data_modeling import (
    DirectRelation,
    DirectRelationReference,
    EdgeId,
    MappedProperty,
    NodeApply,
    NodeId,
)
from cognite.client.data_classes.data_modeling.instances import EdgeApply, NodeOrEdgeData, PropertyValueWrite
from cognite.client.data_classes.data_modeling.views import ViewProperty
from cognite.client.utils._identifier import InstanceId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.migration import (
    AssetCentricId,
    ResourceViewMappingApply,
)
from cognite_toolkit._cdf_tk.utils.collection import flatten_dict_json_path
from cognite_toolkit._cdf_tk.utils.dtype_conversion import (
    asset_centric_convert_to_primary_property,
    convert_to_primary_property,
)
from cognite_toolkit._cdf_tk.utils.useful_types import (
    AssetCentricResourceExtended,
    AssetCentricTypeExtended,
)

from .data_model import INSTANCE_SOURCE_VIEW_ID
from .issues import ConversionIssue, FailedConversion, InvalidPropertyDataType


class DirectRelationCache:
    """Cache for direct relation references to look up target of direct relations.

    This is used when creating direct relations from asset-centric resources to assets, files, and source systems.
    """

    class TableName:
        ASSET_ID = "assetId"
        SOURCE_NAME = "source"
        FILE_ID = "fileId"
        ASSET_EXTERNAL_ID = "assetExternalId"
        FILE_EXTERNAL_ID = "fileExternalId"

    ASSET_ID_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {
        ("timeseries", "assetId"),
        ("file", "assetIds"),
        ("event", "assetIds"),
        ("sequence", "assetId"),
        ("annotation", "data.assetRef.id"),
        ("asset", "parentId"),
    }
    SOURCE_NAME_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {
        ("asset", "source"),
        ("event", "source"),
        ("file", "source"),
    }
    FILE_ID_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {
        ("annotation", "data.fileRef.id"),
        ("annotation", "annotatedResourceId"),
    }
    ASSET_EXTERNAL_ID_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {("annotation", "data.assetRef.externalId")}
    FILE_EXTERNAL_ID_PROPERTIES: ClassVar[Set[tuple[str, str]]] = {("annotation", "data.fileRef.externalId")}

    def __init__(self, client: ToolkitClient) -> None:
        self._client = client
        self._cache_map: dict[
            tuple[str, str] | str, dict[str, DirectRelationReference] | dict[int, DirectRelationReference]
        ] = {}
        # Constructing the cache map to be accessed by both table name and property id
        for table_name, properties in [
            (self.TableName.ASSET_ID, self.ASSET_ID_PROPERTIES),
            (self.TableName.SOURCE_NAME, self.SOURCE_NAME_PROPERTIES),
            (self.TableName.FILE_ID, self.FILE_ID_PROPERTIES),
            (self.TableName.ASSET_EXTERNAL_ID, self.ASSET_EXTERNAL_ID_PROPERTIES),
            (self.TableName.FILE_EXTERNAL_ID, self.FILE_EXTERNAL_ID_PROPERTIES),
        ]:
            cache: dict[str, DirectRelationReference] | dict[int, DirectRelationReference] = {}
            self._cache_map[table_name] = cache
            for key in properties:
                self._cache_map[key] = cache

    def update(self, resources: Iterable[AssetCentricResourceExtended]) -> None:
        """Update the cache with direct relation references for the given asset-centric resources.

        This is used to bulk update the cache for a chunk of resources before converting them to data model instances.
        """
        asset_ids: set[int] = set()
        source_ids: set[str] = set()
        file_ids: set[int] = set()
        asset_external_ids: set[str] = set()
        file_external_ids: set[str] = set()
        for resource in resources:
            if isinstance(resource, Annotation):
                if resource.annotated_resource_type == "file" and resource.annotated_resource_id:
                    file_ids.add(resource.annotated_resource_id)
                if "assetRef" in resource.data:
                    asset_ref = resource.data["assetRef"]
                    if isinstance(asset_id := asset_ref.get("id"), int):
                        asset_ids.add(asset_id)
                    if isinstance(asset_external_id := asset_ref.get("externalId"), str):
                        asset_external_ids.add(asset_external_id)
                if "fileRef" in resource.data:
                    file_ref = resource.data["fileRef"]
                    if isinstance(file_id := file_ref.get("id"), int):
                        file_ids.add(file_id)
                    if isinstance(file_external_id := file_ref.get("externalId"), str):
                        file_external_ids.add(file_external_id)
            elif isinstance(resource, Asset):
                if resource.source:
                    source_ids.add(resource.source)
                if resource.parent_id is not None:
                    asset_ids.add(resource.parent_id)
            elif isinstance(resource, FileMetadata):
                if resource.source:
                    source_ids.add(resource.source)
                if resource.asset_ids:
                    asset_ids.update(resource.asset_ids)
            elif isinstance(resource, Event):
                if resource.source:
                    source_ids.add(resource.source)
                if resource.asset_ids:
                    asset_ids.update(resource.asset_ids)
            elif isinstance(resource, TimeSeries):
                if resource.asset_id is not None:
                    asset_ids.add(resource.asset_id)
        if asset_ids:
            self._update_cache(self._client.migration.lookup.assets(id=list(asset_ids)), self.TableName.ASSET_ID)
        if source_ids:
            # SourceSystems are not cached in the client, so we have to handle the caching ourselves.
            cache = cast(dict[str, DirectRelationReference], self._cache_map[self.TableName.SOURCE_NAME])
            missing = set(source_ids) - set(cache.keys())
            if missing:
                source_systems = self._client.migration.created_source_system.retrieve(list(missing))
                for source_system in source_systems:
                    cache[source_system.source] = source_system.as_direct_relation_reference()
        if file_ids:
            self._update_cache(self._client.migration.lookup.files(id=list(file_ids)), self.TableName.FILE_ID)
        if asset_external_ids:
            self._update_cache(
                self._client.migration.lookup.assets(external_id=list(asset_external_ids)),
                self.TableName.ASSET_EXTERNAL_ID,
            )
        if file_external_ids:
            self._update_cache(
                self._client.migration.lookup.files(external_id=list(file_external_ids)),
                self.TableName.FILE_EXTERNAL_ID,
            )

    def _update_cache(self, instance_id_by_id: dict[int, NodeId] | dict[str, NodeId], table_name: str) -> None:
        cache = self._cache_map[table_name]
        for identifier, instance_id in instance_id_by_id.items():
            cache[identifier] = DirectRelationReference(space=instance_id.space, external_id=instance_id.external_id)  # type: ignore[index]

    def get_cache(
        self, resource_type: AssetCentricTypeExtended, property_id: str
    ) -> Mapping[str | int, DirectRelationReference] | None:
        """Get the cache for the given resource type and property ID."""
        return self._cache_map.get((resource_type, property_id))  # type: ignore[return-value]


def asset_centric_to_dm(
    resource: AssetCentricResourceExtended,
    instance_id: InstanceId,
    view_source: ResourceViewMappingApply,
    view_properties: dict[str, ViewProperty],
    direct_relation_cache: DirectRelationCache,
) -> tuple[NodeApply | EdgeApply | None, ConversionIssue]:
    """Convert an asset-centric resource to a data model instance.

    Args:
        resource (CogniteResource): The asset-centric resource to convert.
        instance_id (NodeId | EdgeApply): The ID of the instance to create or update.
        view_source (ResourceViewMappingApply): The view source defining how to map the resource to the data model.
        view_properties (dict[str, ViewProperty]): The defined properties referenced in the view source mapping.
        direct_relation_cache (DirectRelationCache): Cache for direct relation references.

    Returns:
        tuple[NodeApply | EdgeApply, ConversionIssue]: A tuple containing the converted NodeApply and any ConversionIssue encountered.
    """
    resource_type = _lookup_resource_type(resource)
    dumped = resource.dump()
    try:
        id_ = dumped.pop("id")
    except KeyError as e:
        raise ValueError("Resource must have an 'id' field.") from e
    if not isinstance(id_, int):
        raise TypeError(f"Resource 'id' field must be an int, got {type(id_)}.")
    data_set_id = dumped.pop("dataSetId", None)
    external_id = dumped.pop("externalId", None)

    issue = ConversionIssue(asset_centric_id=AssetCentricId(resource_type, id_=id_), instance_id=instance_id)

    properties = create_properties(
        dumped,
        view_properties,
        view_source.property_mapping,
        resource_type,
        issue=issue,
        direct_relation_cache=direct_relation_cache,
    )
    sources: list[NodeOrEdgeData] = []
    if properties:
        sources.append(NodeOrEdgeData(source=view_source.view_id, properties=properties))

    if resource_type != "annotation":
        instance_source_properties = {
            "resourceType": resource_type,
            "id": id_,
            "dataSetId": data_set_id,
            "classicExternalId": external_id,
        }
        sources.append(NodeOrEdgeData(source=INSTANCE_SOURCE_VIEW_ID, properties=instance_source_properties))

    instance: NodeApply | EdgeApply
    if isinstance(instance_id, EdgeId):
        edge_properties = create_edge_properties(
            dumped, view_source.property_mapping, resource_type, issue, direct_relation_cache, instance_id.space
        )
        if any(key not in edge_properties for key in ("start_node", "end_node", "type")):
            # Failed conversion of edge properties
            return None, issue
        instance = EdgeApply(
            space=instance_id.space,
            external_id=instance_id.external_id,
            sources=sources,
            **edge_properties,  # type: ignore[arg-type]
        )
    elif isinstance(instance_id, NodeId):
        instance = NodeApply(space=instance_id.space, external_id=instance_id.external_id, sources=sources)
    else:
        raise RuntimeError(f"Unexpected instance_id type {type(instance_id)}")

    return instance, issue


def _lookup_resource_type(resource_type: AssetCentricResourceExtended) -> AssetCentricTypeExtended:
    if isinstance(resource_type, Asset):
        return "asset"
    elif isinstance(resource_type, FileMetadata):
        return "file"
    elif isinstance(resource_type, Event):
        return "event"
    elif isinstance(resource_type, TimeSeries):
        return "timeseries"
    elif isinstance(resource_type, Annotation):
        if resource_type.annotated_resource_type == "file" and resource_type.annotation_type in (
            "diagrams.AssetLink",
            "diagrams.FileLink",
        ):
            return "annotation"
    raise ValueError(f"Unsupported resource type: {resource_type}")


def create_properties(
    dumped: dict[str, Any],
    view_properties: dict[str, ViewProperty],
    property_mapping: dict[str, str],
    resource_type: AssetCentricTypeExtended,
    issue: ConversionIssue,
    direct_relation_cache: DirectRelationCache,
) -> dict[str, PropertyValueWrite]:
    """
    Create properties for a data model instance from an asset-centric resource.

    Args:
        dumped: Dict representation of the asset-centric resource.
        view_properties: Defined properties referenced in the view source mapping.
        property_mapping:  Mapping from asset-centric property IDs to data model property IDs.
        resource_type: The type of the asset-centric resource (e.g., "asset", "timeseries").
        issue: ConversionIssue object to log any issues encountered during conversion.
        direct_relation_cache: Cache for direct relation references to look up target of direct relations.

    Returns:
        Dict of property IDs to PropertyValueWrite objects.

    """
    flatten_dump = flatten_dict_json_path(dumped, keep_structured=set(property_mapping.keys()))
    properties: dict[str, PropertyValueWrite] = {}
    ignored_asset_centric_properties: set[str] = set()
    for prop_json_path, prop_id in property_mapping.items():
        if prop_json_path not in flatten_dump:
            continue
        if prop_id not in view_properties:
            continue
        if prop_id in properties:
            ignored_asset_centric_properties.add(prop_json_path)
            continue
        dm_prop = view_properties[prop_id]
        if not isinstance(dm_prop, MappedProperty):
            issue.invalid_instance_property_types.append(
                InvalidPropertyDataType(property_id=prop_id, expected_type=MappedProperty.__name__)
            )
            continue
        try:
            value = asset_centric_convert_to_primary_property(
                flatten_dump[prop_json_path],
                dm_prop.type,
                dm_prop.nullable,
                destination_container_property=(dm_prop.container, dm_prop.container_property_identifier),
                source_property=(resource_type, prop_json_path),
                direct_relation_lookup=direct_relation_cache.get_cache(resource_type, prop_json_path),
            )
        except (ValueError, TypeError, NotImplementedError) as e:
            issue.failed_conversions.append(
                FailedConversion(property_id=prop_json_path, value=flatten_dump[prop_json_path], error=str(e))
            )
            continue
        properties[prop_id] = value

    issue.ignored_asset_centric_properties = sorted(
        (set(flatten_dump.keys()) - set(property_mapping.keys())) | ignored_asset_centric_properties
    )
    issue.missing_asset_centric_properties = sorted(set(property_mapping.keys()) - set(flatten_dump.keys()))
    # Node and edge properties are handled separately
    issue.missing_instance_properties = sorted(
        {
            prop_id
            for prop_id in property_mapping.values()
            if not (prop_id.startswith("edge.") or prop_id.startswith("node."))
        }
        - set(view_properties.keys())
    )
    return properties


def create_edge_properties(
    dumped: dict[str, Any],
    property_mapping: dict[str, str],
    resource_type: AssetCentricTypeExtended,
    issue: ConversionIssue,
    direct_relation_cache: DirectRelationCache,
    default_instance_space: str,
) -> dict[str, DirectRelationReference]:
    flatten_dump = flatten_dict_json_path(dumped)
    edge_properties: dict[str, DirectRelationReference] = {}
    for prop_json_path, prop_id in property_mapping.items():
        if not prop_id.startswith("edge."):
            continue
        if prop_json_path not in flatten_dump:
            continue
        edge_prop_id = prop_id.removeprefix("edge.")
        if edge_prop_id in ("startNode", "endNode", "type"):
            # DirectRelation lookup.
            try:
                value = convert_to_primary_property(
                    flatten_dump[prop_json_path],
                    DirectRelation(),
                    False,
                    direct_relation_lookup=direct_relation_cache.get_cache(resource_type, prop_json_path),
                )
            except (ValueError, TypeError, NotImplementedError) as e:
                issue.failed_conversions.append(
                    FailedConversion(property_id=prop_json_path, value=flatten_dump[prop_json_path], error=str(e))
                )
                continue
        elif edge_prop_id.endswith(".externalId"):
            # Just an external ID string.
            edge_prop_id = edge_prop_id.removesuffix(".externalId")
            value = DirectRelationReference(default_instance_space, str(flatten_dump[prop_json_path]))
        else:
            issue.invalid_instance_property_types.append(
                InvalidPropertyDataType(property_id=prop_id, expected_type="EdgeProperty")
            )
            continue
        # We know that value is DirectRelationReference here
        edge_properties[edge_prop_id.replace("Node", "_node")] = value  # type: ignore[assignment]

    return edge_properties



from collections.abc import Hashable, Iterable, Sequence
from typing import Any, final

from cognite.client.data_classes.capabilities import Capability, StreamsAcl
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.streams import (
    StreamRequest,
    StreamRequestList,
    StreamResponse,
    StreamResponseList,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import StreamYAML
from cognite_toolkit._cdf_tk.utils.http_client import ToolkitAPIError

from .datamodel import ContainerCRUD


@final
class StreamCRUD(ResourceCRUD[str, StreamRequest, StreamResponse, StreamRequestList, StreamResponseList]):
    folder_name = "streams"
    filename_pattern = r"^.*\.Streams$"
    filetypes = frozenset({"yaml", "yml"})
    resource_cls = StreamResponse
    resource_write_cls = StreamRequest
    list_cls = StreamResponseList
    list_write_cls = StreamRequestList
    kind = "Streams"
    yaml_cls = StreamYAML
    dependencies = frozenset({ContainerCRUD})
    _doc_url = "Streams/operation/createStream"
    support_update = False

    @property
    def display_name(self) -> str:
        return "streams"

    @classmethod
    def get_id(cls, item: StreamRequest | StreamResponse | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[StreamRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [StreamsAcl.Action.Read]
            if read_only
            else [StreamsAcl.Action.Read, StreamsAcl.Action.Create, StreamsAcl.Action.Delete]
        )
        return StreamsAcl(actions, StreamsAcl.Scope.All())

    def create(self, items: StreamRequestList) -> StreamResponseList:
        created = self.client.streams.create(list(items))
        return StreamResponseList(created)

    def retrieve(self, ids: SequenceNotStr[str]) -> StreamResponseList:
        retrieved: list[StreamResponse] = []
        for _id in ids:
            try:
                _resp = self.client.streams.retrieve(_id)
            except ToolkitAPIError:
                continue
            if _resp is not None:
                retrieved.append(_resp)
        return StreamResponseList(retrieved)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        count = 0
        for _id in ids:
            try:
                self.client.streams.delete(_id)
            except ToolkitAPIError:
                continue
            count += 1
        return count

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[StreamResponse]:
        if data_set_external_id or space or parent_ids:
            # These filters are not supported for Streams
            return iter([])

        all_streams = self.client.streams.list()
        return iter(all_streams)



from typing import Any

import pytest
from cognite.client.data_classes import DataSet, Event, EventWrite

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.storageio import CanvasIO
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, SuccessResponse


@pytest.fixture
def canvas_event(toolkit_client: ToolkitClient, toolkit_dataset: DataSet) -> Event:
    event = EventWrite(
        external_id="canvas_event_integration_test",
        start_time=1_600_000_000_000,
        end_time=1_600_000_360_000,
        data_set_id=toolkit_dataset.id,
        description="Integration test event for canvas",
        type="canvas_test_event",
        subtype="canvas_test_event",
    )
    if existing := toolkit_client.events.retrieve(external_id=event.external_id):
        return existing
    created_event = toolkit_client.events.create(event)
    return created_event


@pytest.fixture
def canvas_raw_data(canvas_event: Event) -> dict[str, Any]:
    user_id = "ndTFZh9K9-m2W9WBKc30-Q"
    canvas_id = "70c66b27-c16e-4f6c-bc1e-940f086ec2f7"
    event_ref = "167ab784-9f33-4130-831e-91d820359ce9"
    return {
        "canvas": {
            "externalId": canvas_id,
            "instanceType": "node",
            "space": "IndustrialCanvasInstanceSpace",
            "sources": [
                {
                    "source": {
                        "space": "cdf_industrial_canvas",
                        "externalId": "Canvas",
                        "version": "v7",
                        "type": "view",
                    },
                    "properties": {
                        "createdBy": user_id,
                        "isLocked": None,
                        "visibility": "public",
                        "sourceCanvasId": None,
                        "isArchived": None,
                        "context": [{"type": "FILTERS", "payload": {"filters": []}}],
                        "name": "Toolkit Integration Test Canvas",
                        "solutionTags": None,
                        "updatedAt": "2025-11-23T08:09:13.183+00:00",
                        "updatedBy": user_id,
                    },
                }
            ],
        },
        "containerReferences": [
            {
                "externalId": f"{canvas_id}_{event_ref}",
                "instanceType": "node",
                "space": "IndustrialCanvasInstanceSpace",
                "sources": [
                    {
                        "source": {
                            "space": "cdf_industrial_canvas",
                            "externalId": "ContainerReference",
                            "version": "v2",
                            "type": "view",
                        },
                        "properties": {
                            "chartsId": None,
                            "containerReferenceType": "event",
                            "height": 500,
                            "id": event_ref,
                            "label": "canvas_test_event: canvas_event_integration_test",
                            "properties": {"zIndex": 0, "unscaledWidth": 600, "unscaledHeight": 500},
                            "maxHeight": None,
                            "maxWidth": None,
                            "resourceExternalId": canvas_event.external_id,
                            "resourceSubId": None,
                            "width": 600,
                            "x": -10,
                            "y": 418,
                        },
                    }
                ],
            },
        ],
    }


class TestCanvasIO:
    def test_upload_canvas(self, toolkit_client: ToolkitClient, canvas_raw_data: dict[str, Any]) -> None:
        io = CanvasIO(toolkit_client)
        upload_items = io.json_chunk_to_data([("canvas_data.json", canvas_raw_data)])

        assert len(upload_items) == 1

        with HTTPClient(toolkit_client.config) as http_client:
            upload_result = io.upload_items(upload_items, http_client)

        not_success = [message for message in upload_result if not isinstance(message, SuccessResponse)]
        assert len(not_success) == 0, f"Some canvas items failed to upload: {humanize_collection(not_success)}"



import json
import mimetypes
from collections.abc import Iterable, MutableSequence, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import cast

from cognite.client.data_classes import FileMetadata, FileMetadataWrite
from cognite.client.data_classes.data_modeling import NodeId, ViewId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.cruds import FileMetadataCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.fileio import MultiFileReader
from cognite_toolkit._cdf_tk.utils.http_client import (
    DataBodyRequest,
    ErrorDetails,
    FailedResponse,
    FailedResponseItems,
    HTTPClient,
    HTTPMessage,
    ResponseList,
    SimpleBodyRequest,
)
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import Page, UploadableStorageIO, UploadItem
from .selectors import FileContentSelector, FileMetadataTemplateSelector
from .selectors._file_content import FILEPATH, FileDataModelingTemplateSelector

COGNITE_FILE_VIEW = ViewId("cdf_cdm", "CogniteFile", "v1")


@dataclass
class UploadFileContentItem(UploadItem[FileMetadataWrite]):
    file_path: Path
    mime_type: str


class FileContentIO(UploadableStorageIO[FileContentSelector, FileMetadata, FileMetadataWrite]):
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    CHUNK_SIZE = 10
    BASE_SELECTOR = FileContentSelector
    KIND = "FileContent"
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    UPLOAD_ENDPOINT = "/files"

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._crud = FileMetadataCRUD(client, None, None)

    def as_id(self, item: FileMetadata) -> str:
        return item.external_id or str(item.id)

    def stream_data(self, selector: FileContentSelector, limit: int | None = None) -> Iterable[Page]:
        raise NotImplementedError("Download of FileContent is not yet supported")

    def count(self, selector: FileContentSelector) -> int | None:
        return None

    def data_to_json_chunk(
        self, data_chunk: Sequence[FileMetadata], selector: FileContentSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        raise NotImplementedError("Download of FileContent is not yet supported")

    def json_chunk_to_data(self, data_chunk: list[tuple[str, dict[str, JsonVal]]]) -> Sequence[UploadFileContentItem]:
        """Convert a JSON-compatible chunk of data back to a writable Cognite resource list.

        Args:
            data_chunk: A list of tuples, each containing a source ID and a dictionary representing
                the data in a JSON-compatible format.
        Returns:
            A writable Cognite resource list representing the data.
        """
        result: list[UploadFileContentItem] = []
        for source_id, item_json in data_chunk:
            item = self.json_to_resource(item_json)
            filepath = cast(Path, item_json[FILEPATH])
            mime_type, _ = mimetypes.guess_type(filepath)
            # application/octet-stream is the standard fallback for binary data when the type is unknown. (at least Claude thinks so)
            result.append(
                UploadFileContentItem(
                    source_id=source_id,
                    item=item,
                    file_path=filepath,
                    mime_type=mime_type or "application/octet-stream",
                )
            )
        return result

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> FileMetadataWrite:
        return self._crud.load_resource(item_json)

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[FileMetadataWrite]],
        http_client: HTTPClient,
        selector: FileContentSelector | None = None,
    ) -> Sequence[HTTPMessage]:
        results: MutableSequence[HTTPMessage] = []
        if isinstance(selector, FileMetadataTemplateSelector):
            upload_url_getter = self._upload_url_asset_centric
        elif isinstance(selector, FileDataModelingTemplateSelector):
            upload_url_getter = self._upload_url_data_modeling
        elif selector is None:
            raise ToolkitNotImplementedError("Selector must be provided for FileContentIO upload")
        else:
            raise ToolkitNotImplementedError(
                f"Upload for the given selector, {type(selector).__name__}, is not supported for FileContentIO"
            )

        for item in cast(Sequence[UploadFileContentItem], data_chunk):
            if not (upload_url := upload_url_getter(item, http_client, results)):
                continue

            upload_response = http_client.request_with_retries(
                message=DataBodyRequest(
                    endpoint_url=upload_url,
                    method="PUT",
                    content_type=item.mime_type,
                    data_content=item.file_path.read_bytes(),
                )
            )
            results.extend(upload_response.as_item_responses(item.as_id()))
        return results

    def _upload_url_asset_centric(
        self, item: UploadFileContentItem, http_client: HTTPClient, results: MutableSequence[HTTPMessage]
    ) -> str | None:
        responses = http_client.request_with_retries(
            message=SimpleBodyRequest(
                endpoint_url=http_client.config.create_api_url(self.UPLOAD_ENDPOINT),
                method="POST",
                # MyPy does not understand that .dump is valid json
                body_content=item.dump(),  # type: ignore[arg-type]
            )
        )
        return self._parse_upload_link_response(responses, item, results)

    def _upload_url_data_modeling(
        self,
        item: UploadFileContentItem,
        http_client: HTTPClient,
        results: MutableSequence[HTTPMessage],
        created_node: bool = False,
    ) -> str | None:
        """Get upload URL for data modeling file upload.

        We first try to get the upload link assuming the CogniteFile node already exists.
        If we get a "not found" error, we create the CogniteFile node and try again.

        Args:
            item: The upload item containing file metadata.
            http_client: The HTTP client to use for requests.
            results: A mutable sequence to collect HTTP messages and errors.
            created_node: A flag indicating whether the CogniteFile node has already been created.
                This prevents infinite recursion.

        Returns:
            The upload URL as a string, or None if there was an error.

        """
        # We know that instance_id is always set for data modeling uploads
        instance_id = cast(NodeId, item.item.instance_id)
        responses = http_client.request_with_retries(
            message=SimpleBodyRequest(
                endpoint_url=http_client.config.create_api_url("/files/uploadlink"),
                method="POST",
                body_content={"items": [{"instanceId": instance_id.dump(include_instance_type=False)}]},  # type: ignore[dict-item]
            )
        )
        # We know there is only one response since we only requested one upload link
        response = responses[0]
        if isinstance(response, FailedResponse) and response.error.missing and not created_node:
            if self._create_cognite_file_node(instance_id, http_client, item.as_id(), results):
                return self._upload_url_data_modeling(item, http_client, results, created_node=True)
            else:
                return None

        return self._parse_upload_link_response(responses, item, results)

    @classmethod
    def _create_cognite_file_node(
        cls, instance_id: NodeId, http_client: HTTPClient, upload_id: str, results: MutableSequence[HTTPMessage]
    ) -> bool:
        node_creation = http_client.request_with_retries(
            message=SimpleBodyRequest(
                endpoint_url=http_client.config.create_api_url("/models/instances"),
                method="POST",
                body_content={
                    "items": [
                        {
                            "space": instance_id.space,
                            "externalId": instance_id.external_id,
                            "instanceType": "node",
                            # When we create a node with properties in CogniteFile View even with empty properties,
                            # CDF will fill in empty values for all properties defined in the view (note this is only
                            # possible because CogniteFile view has all properties as optional). This includes properties
                            # in the CogniteFile container, which will trigger the file syncer to create a FileMetadata
                            # and link it to the CogniteFile node.
                            "sources": [{"source": COGNITE_FILE_VIEW.dump(include_type=True), "properties": {}}],  # type: ignore[dict-item]
                        }
                    ]
                },
            )
        )
        try:
            _ = node_creation.get_first_body()
        except ValueError:
            results.extend(node_creation.as_item_responses(upload_id))
            return False
        return True

    @classmethod
    def _parse_upload_link_response(
        cls, responses: ResponseList, item: UploadFileContentItem, results: MutableSequence[HTTPMessage]
    ) -> str | None:
        try:
            body = responses.get_first_body()
        except ValueError:
            results.extend(responses.as_item_responses(item.as_id()))
            return None

        if "items" in body and isinstance(body["items"], list) and len(body["items"]) > 0:
            body = body["items"][0]  # type: ignore[assignment]
        try:
            upload_url = cast(str, body["uploadUrl"])
        except (KeyError, IndexError):
            results.append(
                FailedResponseItems(
                    status_code=200,
                    body=json.dumps(body),
                    error=ErrorDetails(code=200, message="Malformed response"),
                    ids=[item.as_id()],
                )
            )
            return None
        return upload_url

    @classmethod
    def read_chunks(
        cls, reader: MultiFileReader, selector: FileContentSelector
    ) -> Iterable[list[tuple[str, dict[str, JsonVal]]]]:
        for chunk in chunker_sequence(reader.input_files, cls.CHUNK_SIZE):
            batch: list[tuple[str, dict[str, JsonVal]]] = []
            for file_path in chunk:
                metadata = selector.create_instance(file_path)
                metadata[FILEPATH] = file_path
                batch.append((str(file_path), metadata))
            yield batch



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.data_classes.canvas import (
    IndustrialCanvas,
    IndustrialCanvasApply,
)
from cognite_toolkit._cdf_tk.client.data_classes.charts import Chart, ChartList, ChartWrite
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.tk_warnings import MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, HTTPMessage, SimpleBodyRequest
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import Page, UploadableStorageIO, UploadItem
from .selectors import (
    AllChartsSelector,
    CanvasExternalIdSelector,
    CanvasSelector,
    ChartExternalIdSelector,
    ChartOwnerSelector,
    ChartSelector,
)


class ChartIO(UploadableStorageIO[ChartSelector, Chart, ChartWrite]):
    KIND = "Charts"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    CHUNK_SIZE = 10
    BASE_SELECTOR = ChartSelector
    UPLOAD_ENDPOINT_TYPE = "app"
    UPLOAD_ENDPOINT_METHOD = "PUT"
    UPLOAD_ENDPOINT = "/storage/charts/charts"

    def as_id(self, item: Chart) -> str:
        return item.external_id

    def stream_data(self, selector: ChartSelector, limit: int | None = None) -> Iterable[Page]:
        selected_charts = self.client.charts.list(visibility="PUBLIC")
        if isinstance(selector, AllChartsSelector):
            ...
        elif isinstance(selector, ChartOwnerSelector):
            selected_charts = ChartList([chart for chart in selected_charts if chart.owner_id == selector.owner_id])
        elif isinstance(selector, ChartExternalIdSelector):
            external_id_set = set(selector.external_ids)
            selected_charts = ChartList([chart for chart in selected_charts if chart.external_id in external_id_set])
        else:
            raise ToolkitNotImplementedError(f"Unsupported selector type {type(selector).__name__!r} for ChartIO")

        if limit is not None:
            selected_charts = ChartList(selected_charts[:limit])
        for chunk in chunker_sequence(selected_charts, self.CHUNK_SIZE):
            yield Page(worker_id="main", items=chunk)

    def count(self, selector: ChartSelector) -> int | None:
        # There is no way to get the count of charts up front.
        return None

    def data_to_json_chunk(
        self, data_chunk: Sequence[Chart], selector: ChartSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        self._populate_timeseries_id_cache(data_chunk)
        return [self._dump_resource(chart) for chart in data_chunk]

    def _populate_timeseries_id_cache(self, data_chunk: Sequence[Chart]) -> None:
        timeseries_ids: set[int] = set()
        for chart in data_chunk:
            for item in chart.data.time_series_collection or []:
                if item.ts_id is not None and item.ts_external_id is None:
                    # We only look-up the internalID if the externalId is missing
                    timeseries_ids.add(item.ts_id)
        if timeseries_ids:
            self.client.lookup.time_series.external_id(list(timeseries_ids))

    def _dump_resource(self, chart: Chart) -> dict[str, JsonVal]:
        dumped = chart.as_write().dump()
        if isinstance(data := dumped.get("data"), dict) and isinstance(
            collection := data.get("timeSeriesCollection"), list
        ):
            for item in collection:
                ts_id = item.pop("tsId", None)
                if ts_id and item.get("tsExternalId") is None:
                    # We only look-up the externalID if it is missing
                    ts_external_id = self.client.lookup.time_series.external_id(ts_id)
                    if ts_external_id is not None:
                        item["tsExternalId"] = ts_external_id
        return dumped

    def json_chunk_to_data(self, data_chunk: list[tuple[str, dict[str, JsonVal]]]) -> Sequence[UploadItem[ChartWrite]]:
        self._populate_timeseries_external_id_cache([item_json for _, item_json in data_chunk])
        return super().json_chunk_to_data(data_chunk)

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> ChartWrite:
        return self._load_resource(item_json)

    def _populate_timeseries_external_id_cache(self, item_jsons: Sequence[dict[str, JsonVal]]) -> None:
        timeseries_external_ids: set[str] = set()
        for item_json in item_jsons:
            if isinstance(data := item_json.get("data"), dict) and isinstance(
                collection := data.get("timeSeriesCollection"), list
            ):
                for item in collection:
                    if not isinstance(item, dict):
                        continue
                    ts_external_id = item.get("tsExternalId")
                    if isinstance(ts_external_id, str):
                        timeseries_external_ids.add(ts_external_id)
        if timeseries_external_ids:
            self.client.lookup.time_series.id(list(timeseries_external_ids))

    def _load_resource(self, item_json: dict[str, JsonVal]) -> ChartWrite:
        if isinstance(data := item_json.get("data"), dict) and isinstance(
            collection := data.get("timeSeriesCollection"), list
        ):
            for item in collection:
                if not isinstance(item, dict):
                    continue
                ts_external_id = item.get("tsExternalId")
                if isinstance(ts_external_id, str) and item.get("tsId") is None:
                    # We only look-up the internalID if it is missing
                    ts_id = self.client.lookup.time_series.id(ts_external_id)
                    if ts_id is not None:
                        item["tsId"] = ts_id
        return ChartWrite._load(item_json)


class CanvasIO(UploadableStorageIO[CanvasSelector, IndustrialCanvas, IndustrialCanvasApply]):
    KIND = "IndustrialCanvas"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    CHUNK_SIZE = 10
    BASE_SELECTOR = CanvasSelector

    def as_id(self, item: IndustrialCanvas) -> str:
        return item.as_id()

    def stream_data(self, selector: CanvasSelector, limit: int | None = None) -> Iterable[Page]:
        if not isinstance(selector, CanvasExternalIdSelector):
            raise ToolkitNotImplementedError(f"Unsupported selector type {type(selector).__name__!r} for CanvasIO")
        canvas_ids = selector.external_ids
        if limit is not None and len(canvas_ids) > limit:
            canvas_ids = canvas_ids[:limit]

        for chunk in chunker_sequence(canvas_ids, self.CHUNK_SIZE):
            items: list[IndustrialCanvas] = []
            for canvas_id in chunk:
                canvas = self.client.canvas.industrial.retrieve(canvas_id)
                if canvas is not None:
                    items.append(canvas)
                else:
                    MediumSeverityWarning("Canvas with external ID {canvas_id!r} not found. Skipping.").print_warning(
                        console=self.client.console
                    )
            yield Page(worker_id="main", items=items)

    def count(self, selector: CanvasSelector) -> int | None:
        if not isinstance(selector, CanvasExternalIdSelector):
            raise ToolkitNotImplementedError(f"Unsupported selector type {type(selector).__name__!r} for CanvasIO")
        return len(selector.external_ids)

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[IndustrialCanvasApply]],
        http_client: HTTPClient,
        selector: CanvasSelector | None = None,
    ) -> Sequence[HTTPMessage]:
        config = http_client.config
        results: list[HTTPMessage] = []
        for item in data_chunk:
            instances = item.item.as_instances()
            responses = http_client.request_with_retries(
                message=SimpleBodyRequest(
                    endpoint_url=config.create_api_url("/models/instances"),
                    method="POST",
                    # MyPy does not understand that .dump is valid json
                    body_content={"items": [instance.dump() for instance in instances]},
                )
            )
            results.extend(responses.as_item_responses(item.source_id))

        return results

    def data_to_json_chunk(
        self, data_chunk: Sequence[IndustrialCanvas], selector: CanvasSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        self._populate_id_cache(data_chunk)
        return [self._dump_resource(canvas) for canvas in data_chunk]

    def _populate_id_cache(self, data_chunk: Sequence[IndustrialCanvas]) -> None:
        """Populate the client's lookup cache with all referenced resources in the canvases."""
        asset_ids: set[int] = set()
        time_series_ids: set[int] = set()
        event_ids: set[int] = set()
        file_ids: set[int] = set()
        for canvas in data_chunk:
            for container_ref in canvas.container_references:
                if container_ref.container_reference_type == "asset":
                    asset_ids.add(container_ref.resource_id)
                elif container_ref.container_reference_type == "timeseries":
                    time_series_ids.add(container_ref.resource_id)
                elif container_ref.container_reference_type == "event":
                    event_ids.add(container_ref.resource_id)
                elif container_ref.container_reference_type == "file":
                    file_ids.add(container_ref.resource_id)
        if asset_ids:
            self.client.lookup.assets.external_id(list(asset_ids))
        if time_series_ids:
            self.client.lookup.time_series.external_id(list(time_series_ids))
        if event_ids:
            self.client.lookup.events.external_id(list(event_ids))
        if file_ids:
            self.client.lookup.files.external_id(list(file_ids))

    def _dump_resource(self, canvas: IndustrialCanvas) -> dict[str, JsonVal]:
        dumped = canvas.as_write().dump()
        references = dumped.get("containerReferences", [])
        if not isinstance(references, list):
            return dumped
        for container_ref in references:
            if not isinstance(container_ref, dict):
                continue
            sources = container_ref.get("sources", [])
            if not isinstance(sources, list) or len(sources) == 0:
                continue
            source = sources[0]
            if not isinstance(source, dict) or "properties" not in source:
                continue
            properties = source["properties"]
            if not isinstance(properties, dict):
                continue
            resource_id = properties.pop("resourceId", None)
            if not isinstance(resource_id, int):
                continue
            reference_type = properties.get("containerReferenceType")
            if reference_type == "asset":
                external_id = self.client.lookup.assets.external_id(resource_id)
            elif reference_type == "timeseries":
                external_id = self.client.lookup.time_series.external_id(resource_id)
            elif reference_type == "event":
                external_id = self.client.lookup.events.external_id(resource_id)
            elif reference_type == "file":
                external_id = self.client.lookup.files.external_id(resource_id)
            else:
                continue
            if external_id is not None:
                properties["resourceExternalId"] = external_id
        return dumped

    def json_chunk_to_data(
        self, data_chunk: list[tuple[str, dict[str, JsonVal]]]
    ) -> Sequence[UploadItem[IndustrialCanvasApply]]:
        self._populate_external_id_cache([item_json for _, item_json in data_chunk])
        return super().json_chunk_to_data(data_chunk)

    def _populate_external_id_cache(self, item_jsons: Sequence[dict[str, JsonVal]]) -> None:
        """Populate the client's lookup cache with all referenced resources in the canvases."""
        asset_external_ids: set[str] = set()
        time_series_external_ids: set[str] = set()
        event_external_ids: set[str] = set()
        file_external_ids: set[str] = set()
        for item_json in item_jsons:
            references = item_json.get("containerReferences", [])
            if not isinstance(references, list):
                continue
            for container_ref in references:
                if not isinstance(container_ref, dict):
                    continue
                sources = container_ref.get("sources", [])
                if not isinstance(sources, list) or len(sources) == 0:
                    continue
                source = sources[0]
                if not isinstance(source, dict) or "properties" not in source:
                    continue
                properties = source["properties"]
                if not isinstance(properties, dict):
                    continue

                resource_external_id = properties.get("resourceExternalId")
                if not isinstance(resource_external_id, str):
                    continue

                reference_type = properties.get("containerReferenceType")
                if reference_type == "asset":
                    asset_external_ids.add(resource_external_id)
                elif reference_type == "timeseries":
                    time_series_external_ids.add(resource_external_id)
                elif reference_type == "event":
                    event_external_ids.add(resource_external_id)
                elif reference_type == "file":
                    file_external_ids.add(resource_external_id)

        if asset_external_ids:
            self.client.lookup.assets.id(list(asset_external_ids))
        if time_series_external_ids:
            self.client.lookup.time_series.id(list(time_series_external_ids))
        if event_external_ids:
            self.client.lookup.events.id(list(event_external_ids))
        if file_external_ids:
            self.client.lookup.files.id(list(file_external_ids))

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> IndustrialCanvasApply:
        return self._load_resource(item_json)

    def _load_resource(self, item_json: dict[str, JsonVal]) -> IndustrialCanvasApply:
        references = item_json.get("containerReferences", [])
        if not isinstance(references, list):
            return IndustrialCanvasApply._load(item_json)
        for container_ref in references:
            if not isinstance(container_ref, dict):
                continue
            sources = container_ref.get("sources", [])
            if not isinstance(sources, list) or len(sources) == 0:
                continue
            source = sources[0]
            if not isinstance(source, dict) or "properties" not in source:
                continue
            properties = source["properties"]
            if not isinstance(properties, dict):
                continue
            resource_external_id = properties.pop("resourceExternalId", None)
            if not isinstance(resource_external_id, str):
                continue
            reference_type = properties.get("containerReferenceType")
            if reference_type == "asset":
                resource_id = self.client.lookup.assets.id(resource_external_id)
            elif reference_type == "timeseries":
                resource_id = self.client.lookup.time_series.id(resource_external_id)
            elif reference_type == "event":
                resource_id = self.client.lookup.events.id(resource_external_id)
            elif reference_type == "file":
                resource_id = self.client.lookup.files.id(resource_external_id)
            else:
                continue
            if resource_id is not None:
                properties["resourceId"] = resource_id

        return IndustrialCanvasApply._load(item_json)



import pytest
import responses
from cognite.client.data_classes import Annotation, AnnotationList

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.storageio import AnnotationIO
from cognite_toolkit._cdf_tk.storageio.selectors import DataSetSelector


class TestAnnotationIO:
    @pytest.mark.usefixtures("disable_pypi_check", "disable_gzip")
    def test_stream_file_annotations(self, toolkit_config: ToolkitClientConfig, rsps: responses.RequestsMock) -> None:
        config = toolkit_config
        client = ToolkitClient(config)
        rsps.add(
            responses.POST,
            config.create_api_url("/files/list"),
            json={
                "items": [
                    {"id": 1, "externalId": "external_1", "name": "file_1"},
                ]
            },
            status=200,
        )
        rsps.add(
            responses.POST,
            config.create_api_url("/files/byids"),
            json={
                "items": [
                    {"id": 1, "externalId": "external_1", "name": "file_1"},
                    {"id": 2, "externalId": "external_2", "name": "file_2"},
                ]
            },
            status=200,
        )
        rsps.add(
            responses.POST,
            config.create_api_url("/assets/byids"),
            json={
                "items": [
                    {"id": 3, "externalId": "external_3", "name": "asset_3"},
                ]
            },
            status=200,
        )
        annotations = AnnotationList(
            [
                Annotation(
                    annotation_type="diagrams.FileLink",
                    status="approved",
                    creating_app="test_app",
                    creating_app_version="v1",
                    annotated_resource_id=1,
                    annotated_resource_type="file",
                    creating_user="test_user",
                    data={
                        "fileRef": {
                            "id": 2,
                        },
                        "textRegion": {
                            "xMin": 0,
                            "xMax": 100,
                            "yMin": 0,
                            "yMax": 100,
                        },
                    },
                ),
                Annotation(
                    annotation_type="diagrams.AssetLink",
                    status="approved",
                    creating_app="test_app",
                    creating_app_version="v1",
                    annotated_resource_id=1,
                    annotated_resource_type="file",
                    creating_user="test_user",
                    data={
                        "assetRef": {"id": 3},
                        "textRegion": {
                            "xMin": 10,
                            "xMax": 110,
                            "yMin": 10,
                            "yMax": 110,
                        },
                    },
                ),
            ]
        )
        for annotation in annotations:
            rsps.add(
                responses.POST,
                config.create_api_url("/annotations/list"),
                json={"items": [annotation.dump()]},
                status=200,
            )

        annotation_io = AnnotationIO(client)
        selector = DataSetSelector(kind="FileMetadata", data_set_external_id="test_data_set")

        pages = list(annotation_io.stream_data(selector))
        assert len(pages) == 2
        total_items = sum(len(page.items) for page in pages)
        assert total_items == 2
        assert annotation_io.count(selector) is None

        json_data = [data for page in pages for data in annotation_io.data_to_json_chunk(page.items)]
        assert len(json_data) == 2
        assert json_data == [
            {
                "annotatedResourceExternalId": "external_1",
                "annotatedResourceType": "file",
                "annotationType": "diagrams.FileLink",
                "creatingApp": "test_app",
                "creatingAppVersion": "v1",
                "creatingUser": "test_user",
                "data": {
                    "fileRef": {"externalId": "external_2"},
                    "textRegion": {"xMax": 100, "xMin": 0, "yMax": 100, "yMin": 0},
                },
                "status": "approved",
            },
            {
                "annotatedResourceExternalId": "external_1",
                "annotatedResourceType": "file",
                "annotationType": "diagrams.AssetLink",
                "creatingApp": "test_app",
                "creatingAppVersion": "v1",
                "creatingUser": "test_user",
                "data": {
                    "assetRef": {"externalId": "external_3"},
                    "textRegion": {"xMax": 110, "xMin": 10, "yMax": 110, "yMin": 10},
                },
                "status": "approved",
            },
        ]



import time
from pathlib import Path

from cognite.client.data_classes import FileMetadataWrite
from cognite.client.data_classes.data_modeling import NodeId, Space

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.storageio import FileContentIO
from cognite_toolkit._cdf_tk.storageio._file_content import UploadFileContentItem
from cognite_toolkit._cdf_tk.storageio.selectors import (
    FileDataModelingTemplate,
    FileDataModelingTemplateSelector,
    FileMetadataTemplate,
    FileMetadataTemplateSelector,
)
from cognite_toolkit._cdf_tk.storageio.selectors._file_content import FILENAME_VARIABLE, TemplateNodeId
from cognite_toolkit._cdf_tk.utils.fileio import MultiFileReader
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient
from tests.test_integration.constants import RUN_UNIQUE_ID


class TestFileContentIO:
    def test_upload_file(self, toolkit_client: ToolkitClient, tmp_path: Path) -> None:
        my_text_file = tmp_path / "my_file.txt"
        my_text_file.write_text("This is some test content.", encoding="utf-8")
        metadata = FileMetadataWrite(
            name="my_file.txt",
            external_id=f"test_upload_file_001_{RUN_UNIQUE_ID}",
            source="TestUpload",
        )

        io = FileContentIO(toolkit_client)
        try:
            with HTTPClient(toolkit_client.config) as http_client:
                io.upload_items(
                    [
                        UploadFileContentItem(
                            source_id=my_text_file.name,
                            item=metadata,
                            file_path=my_text_file,
                            mime_type="text/plain",
                        )
                    ],
                    http_client,
                    FileMetadataTemplateSelector(
                        file_directory=Path("does not matter"),
                        template=FileMetadataTemplate(name=FILENAME_VARIABLE, external_id=FILENAME_VARIABLE),
                    ),
                )
            # Verify upload
            uploaded_file = toolkit_client.files.retrieve(external_id=metadata.external_id)
            assert uploaded_file is not None
            assert uploaded_file.name == "my_file.txt"
            assert uploaded_file.uploaded is True
        finally:
            # Clean up
            toolkit_client.files.delete(external_id=metadata.external_id, ignore_unknown_ids=True)

    def test_read_deserialize_upload_dm_file(
        self, toolkit_client: ToolkitClient, tmp_path: Path, toolkit_space: Space
    ) -> None:
        external_id = f"test_upload_dm_file_001_{RUN_UNIQUE_ID}.txt"
        my_text_file = tmp_path / external_id
        my_text_file.write_text("This is some test content for data modeling.", encoding="utf-8")
        instance_id = NodeId(space=toolkit_space.space, external_id=external_id)
        selector = FileDataModelingTemplateSelector(
            file_directory=tmp_path,
            template=FileDataModelingTemplate(
                instance_id=TemplateNodeId(
                    space=toolkit_space.space,
                    external_id=FILENAME_VARIABLE,
                )
            ),
        )
        io = FileContentIO(toolkit_client)
        reader = MultiFileReader([my_text_file])

        read_chunks = list(io.read_chunks(reader, selector))
        assert len(read_chunks) == 1
        read_chunk = read_chunks[0]

        upload_content = io.json_chunk_to_data(read_chunk)
        assert len(upload_content) == 1

        try:
            with HTTPClient(toolkit_client.config) as http_client:
                io.upload_items(upload_content, http_client, selector)

            t0 = time.perf_counter()
            while True:
                uploaded_file = toolkit_client.files.retrieve(instance_id=instance_id)
                assert uploaded_file is not None
                if uploaded_file.uploaded is True:
                    break
                # The file syncer may take some time to update CogniteFile -> FileMetadata uploaded status
                if time.perf_counter() - t0 > 30:
                    raise AssertionError("Timeout waiting for file to be uploaded.")
                time.sleep(1)
        finally:
            # Clean up
            toolkit_client.data_modeling.instances.delete(nodes=instance_id)



from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

import pytest
from _pytest.monkeypatch import MonkeyPatch

from cognite_toolkit import _version
from cognite_toolkit._cdf_tk.cdf_toml import CDFToml, CLIConfig, ModulesConfig
from cognite_toolkit._cdf_tk.commands.init import InitCommand
from cognite_toolkit._cdf_tk.commands.repo import RepoCommand
from tests.test_unit.utils import MockQuestionary


# Reset singleton before each test to ensure test isolation
@pytest.fixture(autouse=True)
def reset_cdf_toml_singleton():
    """Reset CDFToml singleton before and after each test to ensure test isolation."""
    global _CDF_TOML
    _CDF_TOML = None
    yield
    _CDF_TOML = None  # Clean up after test as well


class TestInitCommand:
    @staticmethod
    def _mock_cdf_toml_load_non_loaded(cls, cwd=None, use_singleton=True):
        """Mock CDFToml.load to return a non-loaded instance (as if cdf.toml doesn't exist)."""
        cwd = cwd or Path.cwd()
        return CDFToml(
            cdf=CLIConfig(cwd),
            modules=ModulesConfig.load({"version": _version.__version__}),
            alpha_flags={},
            plugins={},
            libraries={},
            is_loaded_from_file=False,
        )

    def test_cdf_toml_is_always_created(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test that cdf.toml is always created when running init."""
        cdf_toml_path = tmp_path / "cdf.toml"
        if cdf_toml_path.exists():
            cdf_toml_path.unlink()

        with monkeypatch.context() as m:
            m.chdir(tmp_path)

            user_input = [
                "initToml",  # Select: "Create toml file"
                "",  # Text: organization_dir prompt (empty = current directory)
                "__exit__",  # Select: Exit
            ]
            with (
                patch.object(CDFToml, "load", classmethod(self._mock_cdf_toml_load_non_loaded)),
                MockQuestionary(
                    [
                        "cognite_toolkit._cdf_tk.commands.init",
                        "cognite_toolkit._cdf_tk.commands.modules",
                    ],
                    monkeypatch,
                    user_input,
                ),
            ):
                cmd = InitCommand(print_warning=False, skip_tracking=True)
                cmd.execute(emulate_dot_seven=True)

        assert cdf_toml_path.exists(), "cdf.toml should be created"
        assert CDFToml.load(tmp_path, use_singleton=False).is_loaded_from_file

    def test_github_folder_created_when_selecting_repo_and_github(
        self, tmp_path: Path, monkeypatch: MonkeyPatch
    ) -> None:
        """Test that selecting repo and GitHub creates a .github folder."""
        with monkeypatch.context() as m:
            m.chdir(tmp_path)

            user_input = [
                "initToml",  # Select: "Create toml file"
                "",  # Text: organization_dir prompt (empty = current directory)
                "initRepo",  # Select: "Git repository"
                "GitHub",  # Select: GitHub hosting (from repo module)
                "__exit__",  # Select: Exit
            ]
            with (
                patch.object(CDFToml, "load", classmethod(self._mock_cdf_toml_load_non_loaded)),
                MockQuestionary(
                    [
                        "cognite_toolkit._cdf_tk.commands.init",
                        "cognite_toolkit._cdf_tk.commands.modules",
                        "cognite_toolkit._cdf_tk.commands.repo",
                    ],
                    monkeypatch,
                    user_input,
                ),
            ):
                # Patch _init_repo to use RepoCommand with skip_git_verify
                def mock_init_repo(self, dry_run: bool = False) -> None:
                    if dry_run:
                        return
                    repo_command = RepoCommand(skip_git_verify=True, print_warning=False, skip_tracking=True)
                    repo_command.run(lambda: repo_command.init(cwd=Path.cwd(), host=None, verbose=False))

                with patch.object(InitCommand, "_init_repo", mock_init_repo):
                    cmd = InitCommand(print_warning=False, skip_tracking=True)
                    cmd.execute(emulate_dot_seven=True)

        # Verify .github folder was created
        github_folder = tmp_path / ".github" / "workflows"
        assert github_folder.exists(), ".github/workflows folder should be created"
        assert (github_folder / "deploy.yaml").exists() and (github_folder / "dry-run.yaml").exists()

    def test_exit_actually_exits(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test that selecting exit actually exits the command."""
        with monkeypatch.context() as m:
            m.chdir(tmp_path)
            user_input = ["__exit__"]
            with MockQuestionary(
                ["cognite_toolkit._cdf_tk.commands.init"],
                monkeypatch,
                user_input,
            ):
                cmd = InitCommand(print_warning=False, skip_tracking=True)
                cmd.execute(emulate_dot_seven=True)
